"""The :py:mod:`ssb_timeseries.sample_data` module provides tools for creating sample timeseries data. These are convenience functions for tests and demos."""

from __future__ import annotations

import itertools
from datetime import datetime
from datetime import timedelta
from functools import partial
from typing import Any

import narwhals as nw
import numpy as np
from dateutil.relativedelta import relativedelta
from dateutil.rrule import DAILY
from dateutil.rrule import HOURLY
from dateutil.rrule import MINUTELY
from dateutil.rrule import MONTHLY
from dateutil.rrule import SECONDLY
from dateutil.rrule import WEEKLY
from dateutil.rrule import YEARLY
from dateutil.rrule import rrule

from ssb_timeseries.dates import date_round
from ssb_timeseries.dates import ensure_datetime

# mypy: disable-error-code="arg-type, type-arg, import-untyped, unreachable, attr-defined"


def series_names(*args: dict | str | list[str] | tuple, **kwargs: str) -> list[str]:
    """Return all permutations of the elements of multiple groups of strings.

    Args:
        *args (str | list | tuple | dict): Each arg in args should be a collection of names to be combined with the other.

        **kwargs (str): One option: 'separator' defines a character sequence inserted between name elements. Defaults to '_'.

    The choice of '_' as default separator is not arbitrary: Some functionality,notably Dataset.vectors(), use series names to create Python variables.
    A default separator that is valid in a variable name simnplifies that.

    Returns:
        list[str]: List of names to be used as series names.

    Raises:
        ValueError: If an argument of an invalid type is passed.

    """
    # The real dataseries uses the sign . as separator
    # "Real dataseries" = "The most used naming convention for series in the legacy FAME databases of Statistics Norway"
    separator = kwargs.get("separator", "_")

    if len(args) == 1 and isinstance(args[0], dict):
        return list(args[0].values())

    final_args = []

    for arg in args:
        if arg is None:
            final_args.append([""])
        elif isinstance(arg, str):
            final_args.append([arg])
        elif isinstance(arg, list):
            final_args.append(arg)
        elif isinstance(arg, tuple):
            final_args.append(list(arg))
        elif isinstance(arg, dict):
            final_args.append(list(arg.values()))
        else:
            raise ValueError(f"Invalid argument type: {type(arg)}")

    names = [
        separator.join(combination) for combination in itertools.product(*final_args)
    ]

    return names


def create_df(
    *lists: dict | list[str] | tuple | str,
    start_date: datetime | str = "",
    end_date: datetime | str = "",
    freq: str = "D",
    interval: int = 1,
    separator: str = "_",
    midpoint: int | float = 100,
    variance: int | float = 10,
    temporality: str = "AT",
    decimals: int = 0,
    implementation: str = "pandas",
) -> Any:
    """Generate sample data for specified date range and permutations over lists.

    Args:
        start_date (datetime): The start date of the date range. Optional, default is today - 365 days.
        end_date (datetime): The end date of the date range. Optional, default is today.
        *lists (list[str]): Lists of values to generate combinations from.
        freq (str): The frequency of date generation.
            'Y' for yearly at last day of year,
            'YS' for yearly at first day of year,
            'M' for monthly at last day of month,
            'MS' for monthly at first day of month,
            'W' for weekly on Sundays,
            'D' for daily,
            'H' for hourly,
            'T' for minutely,
            'S' for secondly,
            etc.
            Optional, default is 'D'.
        interval: The interval between dates. ; optionalDefault is 1.
        separator: The separator used to join combinations. Optional, default is '_'.
        midpoint: The midpoint value for generating random data. Optional, default is 100.
        variance: The variance value for generating random data. Optional, default is 10.
        temporality: The temporality of the data. Default is 'AT'.
        decimals: The number of decimal places to round to. Optional, default is 0.
        implementation: Narwhals supported dataframe library or object type.

    Returns:
        A DataFrame or similar object (Numpy array, Arrow table, dict) containing sample data.

    Example:
    ```
    # Generate sample data with no specified start or end date (defaults to +/- infinity)
    sample_data = generate_sample_df(List1, List2, freq='D')
    ```
    """
    if not start_date:
        start_date = date_round(datetime.now()) - timedelta(days=364)
    if not end_date:
        end_date = date_round(datetime.now())

    series = series_names(*lists, separator=separator)
    dates = date_ranges(
        start_date=ensure_datetime(start_date),
        end_date=ensure_datetime(end_date),
        freq=freq,
        interval=interval,
        temporality=temporality,
    )

    rows = len(dates.get("valid_at", dates.get("valid_from")))
    cols = len(series)
    numbers = random_numbers(
        rows, cols, midpoint=midpoint, variance=variance, decimals=decimals
    )
    data_dict = {**dates, **{name: numbers[:, i] for i, name in enumerate(series)}}

    if implementation == "dict":
        return data_dict
    else:
        nw_df = nw.from_dict(data_dict, backend=implementation)
        match implementation.lower():
            case "pyarrow" | "arrow" | "pa":
                return nw_df.to_arrow()
            case "numpy" | "np":
                return nw_df.to_numpy()
            case "polars" | "pl":
                return nw_df.to_polars()
            case "narwhals" | "nw":
                return nw_df
            case "pandas" | "pd" | _:
                return nw_df.to_pandas()


def date_ranges(
    start_date: datetime | str,
    end_date: datetime | str,
    freq: str,
    interval: int = 1,
    temporality: str = "AT",
) -> dict[str, list[datetime]]:
    """Generate a list of dates with a specified frequency."""
    freq_map = {
        "Y": YEARLY,
        "YS": YEARLY,
        "YE": YEARLY,
        "M": MONTHLY,
        "MS": MONTHLY,
        "ME": MONTHLY,
        "W": WEEKLY,
        "D": DAILY,
        "H": HOURLY,
        "T": MINUTELY,
        "S": SECONDLY,
    }
    if freq[0] == "Q":
        interval *= 3
        freq = freq.replace("Q", "M")

    if freq[0] == "Y":
        bymonth = 1
        bymonthday = 1
        if freq[-1] == ("E"):
            bymonthday = -1
    elif freq in ("M", "MS"):
        bymonth = None
        bymonthday = 1
    elif freq in ("ME"):
        bymonth = None
        bymonthday = -1
    else:
        bymonth = None
        bymonthday = None

    r = partial(
        rrule,
        freq=freq_map[freq.upper()],
        interval=interval,
        bymonth=bymonth,
        bymonthday=bymonthday,
    )
    dt_start = ensure_datetime(start_date)
    dt_end = ensure_datetime(end_date)
    d = r(
        dtstart=dt_start,
        until=dt_end,
    )
    d_list = list(d)
    if str(temporality) == "AT":
        return {"valid_at": d_list}
    elif str(temporality) == "FROM_TO":
        delta = relativedelta(d_list[1], d_list[0])
        d_to = r(dtstart=dt_start + delta, until=dt_end + delta)
        return {"valid_from": d_list, "valid_to": list(d_to)}
    else:
        raise ValueError(f"Unhandled temporality: {temporality}")


def random_numbers(
    rows: int,
    cols: int,
    decimals: int = 0,
    midpoint: int | float = 100,
    variance: int | float = 10,
) -> np.ndarray:
    """Generate sample dataframe of specified dimensions."""
    generator = np.random.default_rng()
    random_matrix = generator.standard_normal(size=(rows, cols))
    return midpoint + variance * random_matrix.round(decimals)


def xyz_at(implementation: str = "pandas") -> Any:
    """Return a :py:class:`Temporality.AT` compliant dataframe with a year of monthly data for series 'x', 'y' and 'z'."""
    df = create_df(
        ["x", "y", "z"],
        start_date="2022-01-01",
        end_date="2022-12-31",
        freq="MS",
        temporality="AT",
        implementation=implementation,
    )
    return df


def xyz_from_to(implementation: str = "pandas") -> Any:
    """Return a :py:class:`Temporality.FROM_TO` compliant dataframe with a year of monthly data for series 'x', 'y' and 'z'."""
    df = create_df(
        ["x", "y", "z"],
        start_date="2022-01-01",
        end_date="2022-12-31",
        freq="MS",
        temporality="FROM_TO",
        implementation=implementation,
    )
    return df



"""Logging for time series library.

Logging every dataset read and write operation is useful for process monitoring and provides set level data lineage.
Writes are natural drivers for an automated event based workflow.

It depends on the system context which role the time series library should play.
Workflow control and monitoring should generally be managed by other components,
but the time series system should play nicely with whatever these might be.
In practical terms, that boils down to structured logging.
That in turn allow generating events from the log information.
In some system setups a backend database or an API could take care of logging and event generation.
However, the timeseries library could also be used with only a file system or object storage as backend.
In that case, the library is possibly the most central point close to the reads and writes.

Since the library requires proper configuration anyway, the natural solution is to let the configuration control the logging behaviour as well.
This also leave it up to the application to make decisions on logging;
in line with the general recommendation for code libraries, https://docs.python.org/3.10/howto/logging.html#library-config.

Legacy: If the configuration contains a non empty `log_file` field, logs are written there (and to console).

If `logging` field is specified, it takes presedence, and is used with `logging.config.dictConfig`.
If both are empty, or dictConfig does not specify any log handlers, logging should be disabled.

TO DO: Add support for a 'workflow' log handler that puts all entries with log level INFO onto a queue.
(A default/option for logging configuration may be sufficient?)

TODO: Logs should provide lineage at the dataset level, identifying which named processes reads and writes data.
That requires log messages to include process names, UUIDs or other identifiers.
The time series library may be able to get this information (from the scope of the calling code) by stack inspection.
Otherwise, it must be passed in as parameters to read/write functions.
"""

from __future__ import annotations

# Consider whether:
#  - logging could/should(?) have its own Dapla library.
#  - "everything" could/should be done in ssb_timeseries/__init__.py?
import functools
import logging
from collections.abc import Callable
from datetime import datetime
from logging.config import dictConfig
from typing import Any

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self  # noqa: UP035 #backport to 3.10

# import ssb_timeseries as ts

# mypy: disable-error-code="operator, no-untyped-def, return-value, import-untyped, arg-type"

_STRING_FORMAT: str = "%(name)s | %(levelname)s | %(asctime)s | %(message)s \n"
_JSON_FORMAT: str = '{"name": "%(name)s"; "level": %(levelname)s; "timestamp": %(asctime)s; "message": "%(message)s" }'


def console_handler(
    format_string: str = _STRING_FORMAT,
) -> logging.StreamHandler:
    """Create handler for logging to console."""
    log_handler = logging.StreamHandler()
    log_handler.setFormatter(logging.Formatter(format_string))
    log_handler.setLevel("INFO")
    return log_handler


def file_handler(file: str, format_string: str = _JSON_FORMAT) -> logging.FileHandler:
    """Create handler for logging to file."""
    log_handler = logging.FileHandler(file)
    log_handler.setFormatter(logging.Formatter(format_string))
    log_handler.setLevel("INFO")
    return log_handler


# def coogle_cloud_log_handler():
# automatic cloud logging config
# import google.cloud.logging
# client = google.cloud.logging.Client()
# client.setup_logging()

# https://docs.python.org/3/howto/logging-cookbook.html#filters-contextual
# Consider custom adapter for adding context?


# suggested convenience functionality intended for processes of the statistics production
# ... should be moved out of timeseries library?
class EnterExitLog:
    """Class supporting decorator to log on enter and exit."""

    def __init__(self, name: str, logger: logging.Logger) -> None:
        """Enter/exit template for workflow process."""
        self.name = name

    def __enter__(self) -> Self:
        """Before each workflow process step, do this."""
        self.init_time = datetime.now()
        logger.info(f"START: {self.name}.")
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:  # noqa ANN001
        """After each workflow process step, do this."""
        self.end_time = datetime.now()
        self.elapsed_time = self.end_time - self.init_time
        if not exc_type:
            logger.info(
                "FINISH: %s completed successfully. Execution time: %s seconds.",
                self.name,
                self.elapsed_time,
            )
        else:
            logger.error(
                "FAILED: %s failed. Execution time: %s seconds.",
                self.name,
                self.elapsed_time,
            )


def log_start_stop(func: Callable) -> Callable:
    """Log start and stop of decorated function."""
    # nosonar  TODO: generalise: pass in functions to enter/exit?

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        name = kwargs.pop("logger", __package__)
        logger = logging.getLogger(name)
        with EnterExitLog(name=func.__name__, logger=logger):
            out = func(*args, **kwargs)

        return out

    return wrapper


class LoggerNameNotDefined(Exception):
    """The provided logger name is not defined in the accompanying configuration."""

    ...


def set_up_logging_according_to_config(
    name: str,
    log_config: dict[str, Any],
) -> logging.Logger:
    """Set up logging according to configuration.

    Normally, this happens only in :py:mod:`ssb_timeseries.__init__.py`, which is run when the package is first imported.
    (For later imports, a cached instance will be retrieved.)
    """
    logger = logging.getLogger(name)
    # loop instead of  logger.handlers.clear() is more robust?
    # (suggested as fix for ResourceWarnings)
    for handler in logger.handlers[:]:
        handler.close()
        logger.removeHandler(handler)

    logger.propagate = False
    logger.setLevel("INFO")

    if isinstance(log_config, dict) and "version" in log_config.keys():
        if name not in log_config["loggers"]:
            raise LoggerNameNotDefined(f"Logger '{name}' not defined in {log_config}")

        dictConfig(log_config)
        logger = logging.getLogger(name)
    else:
        # if log_config == {}: #TODO: check what happens for valid dictConfig if all handlers are disabled
        logger.addHandler(logging.NullHandler())

        # dictConfig(log_config)
        # legacy set up
        # logger.addHandler(console_handler())
        # logger.addHandler(file_handler(log_file))

    return logging.getLogger(name)


if __name__ == "__main__":
    ...
else:
    logger = logging.getLogger(__package__)
    # from ssb_timeseries.config import CONFIG
    # logger = set_up_logging_according_to_config(__package__,CONFIG)



"""The main purpose of the :py:mod:`ssb_timeseries.fs` module is to allow file based IO that works on both a local file system and Google Cloud Storage."""

from __future__ import annotations

import functools
import glob
import json
import os
import shutil
from _collections_abc import Callable
from pathlib import Path

import narwhals
import pyarrow
import pyarrow.dataset
import pyarrow.parquet as pq
import tomli
import tomli_w
from dapla import FileClient
from narwhals.typing import IntoFrame

from ..dataframes import to_arrow
from ..types import F
from ..types import PathStr

# mypy: disable-error-code="arg-type, type-arg, no-any-return, no-untyped-def, import-untyped, attr-defined, type-var, index, return-value"


def path_to_str(path: PathStr) -> PathStr:
    """Normalise as strings.

    This is a trick to make automated tests pass on Windows.
    """
    return str(Path(path)).replace("gs:/", "gs://")


def wrap_return_as_str(func: F) -> F:
    """Decorator to normalise outputs using path_to_str()."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        out = func(*args, **kwargs)
        return path_to_str(out)

    return wrapper


def remove_prefix(path: PathStr) -> str:
    """Helper function to compensate for some os.* functions shorten gs://<path> to gs:/<path>."""
    return str(path).replace("//", "/").replace("gs:/", "")


def is_gcs(path: PathStr) -> bool:
    """Check if path is on GCS."""
    return str(path).startswith("gs:/")


def is_local(path: PathStr) -> bool:
    """Check if path is local."""
    return not str(path).startswith("gs:/")


def fs_type(path: PathStr) -> str:
    """Check filesystem type (local or GCS) for a given path."""
    out = ""
    types = {"gcs": is_gcs(path), "local": is_local(path)}
    out = list(types.keys())[list(types.values()).index(True)]
    return out


def exists(path: PathStr) -> bool:
    """Check if a given (local or GCS) path exists."""
    # OSError: [Errno 36] File name too long
    # New (temporary) dataset names generated by functions can become very long.
    # The above error then occurs when metadata maintenance functions check if files exist.
    # Hence the need toi control for filenames
    p = Path(path)
    MAX_FILENAME_LENGTH = 254
    if len(p.name) > MAX_FILENAME_LENGTH:
        path = p.parent / p.name[:MAX_FILENAME_LENGTH]

    if not path:
        return False
    elif is_gcs(path):
        fs = FileClient.get_gcs_file_system()
        return fs.exists(path)
    else:
        return Path(path).exists()


@wrap_return_as_str
def existing_subpath(path: PathStr) -> PathStr:
    """Return the existing part of a path on local or GCS file system."""
    if Path(path).exists():
        return str(path)
    else:
        p = Path(path).parent
        while not p.exists():
            p = Path(p).parent
        return p


def touch(path: PathStr) -> PathStr:
    """Touch file regardless of wether the filesystem is local or GCS; return path."""
    if is_gcs(path):
        fs = FileClient.get_gcs_file_system()
        fs.touch(path)
    else:
        mk_parent_dir(path)
        Path(path).touch()
    return path


@wrap_return_as_str
def path(*args: PathStr) -> str:
    """Join args to form path. Make sure that gcs paths are begins with double slash: gs://..."""
    p = Path(args[0]).joinpath(*args[1:])
    return p


def mkdir(path: PathStr) -> None:
    """Make directory regardless of filesystem is local or GCS."""
    # not good enough .. it is hard to distinguish between dirs and files that do not exist yet
    if is_local(path):
        Path(path).mkdir(parents=True, exist_ok=True)
    else:
        ...


def mk_parent_dir(path: PathStr) -> None:
    """Ensure a parent directory exists. ... regardless of wether fielsystem is local or GCS."""
    # wanted a mkdir that could work seamlessly with both file and directory paths,
    # but it is hard to distinguish between dirs and files that do not exist yet
    # --> use this to create parent directory for files, mkdir() when the last part of path is a directory
    if is_local(path):
        Path(path).parent.mkdir(parents=True, exist_ok=True)
    else:
        ...


def file_count(path: PathStr, create: bool = False) -> int:
    """Count files in path. Should work regardless of wether source and target location is local fs or GCS to local."""
    return len(ls(path, create=create))


def ls(path: str, pattern: str = "*", create: bool = False) -> list[str]:
    """List files. Should work regardless of wether the filesystem is local or GCS."""
    search = os.path.join(path, pattern)
    if is_gcs(path):
        fs = FileClient.get_gcs_file_system()
        return fs.glob(search)
    else:
        if create:
            mkdir(path)
        return glob.glob(search)


def cp(from_path: PathStr, to_path: PathStr) -> None:
    """Copy file from one location to another.

    This function handles copying files between local and GCS paths,
    automatically selecting the correct backend.

    Args:
        from_path: The path to the source file.
        to_path: The path to the destination file.
    """
    from_type = fs_type(from_path)
    to_type = fs_type(to_path)
    if is_gcs(from_path) | is_gcs(to_path):
        fs = FileClient.get_gcs_file_system()
    if is_local(to_path):
        mk_parent_dir(to_path)

    match (from_type, to_type):
        case ("local", "local"):
            shutil.copy2(from_path, to_path)
        case ("local", "gcs"):
            fs.put(from_path, to_path)
        case ("gcs", "local"):
            fs.get(from_path, to_path)
        case ("gcs", "gcs"):
            fs.copy(from_path, to_path)


def mv(from_path: PathStr, to_path: PathStr) -> None:
    """Move file from one location to another.

    This function handles moving files between local and GCS paths,
    automatically selecting the correct backend.

    Args:
        from_path: The path to the source file.
        to_path: The path to the destination file.
    """
    from_type = fs_type(from_path)
    to_type = fs_type(to_path)

    if is_gcs(from_path) | is_gcs(to_path):
        fs = FileClient.get_gcs_file_system()
    if is_local(to_path):
        mk_parent_dir(to_path)

    match (from_type, to_type):
        case ("local", "local"):
            shutil.move(from_path, to_path)
        case ("local", "gcs"):
            fs.put(from_path, to_path)
        case ("gcs", "local"):
            fs.get(from_path, to_path)
        case ("gcs", "gcs"):
            fs.move(from_path, to_path)


def rm(path: PathStr) -> None:
    """Remove a file from either the local filesystem or GCS.

    This function is non-recursive. For a recursive variant, see rmtree().

    Args:
        path: The path to the file to be removed.
    """
    if is_gcs(path):
        fs = FileClient.get_gcs_file_system()
        fs.rm(path)
    else:
        os.remove(path)


def rmtree(
    path: str,
) -> None:
    """Recursively remove a directory and all its subdirectories and files regardless of local or GCS filesystem."""
    if is_gcs(path):
        ...
        # TO DO: implement this (but recursive)
        # fs = FileClient.get_gcs_file_system()
        # fs.rm(path)
    else:
        shutil.rmtree(path)


@wrap_return_as_str
def same_path(*args) -> PathStr:
    """Return common part of path, for two or more files. Files must be on same file system, but the file system can be either local or GCS."""
    # TO DO: add support for Windows style paths?
    # ... regex along the lines of: [A-Z\:|\\\\]
    paths = [a.replace("gs:/", "") for a in args]
    return os.path.commonpath(paths)


def find(
    search_path: PathStr,
    equals: str = "",
    contains: str = "",
    pattern: str = "",
    search_sub_dirs: bool = True,
    full_path: bool = False,
    replace_root: bool = False,
) -> list[str]:
    """Find files and subdirectories with names matching pattern. Should work for both local and GCS filesystems."""
    if contains:
        pattern = f"*{pattern}*"
    elif equals:
        pattern = equals
    elif not pattern:
        pattern = "*"

    if search_sub_dirs:
        search_str = path(search_path, "*", pattern)
    else:
        search_str = path(search_path, pattern)

    if is_gcs(path):
        fs = FileClient.get_gcs_file_system()
        found = fs.glob(search_str)
    else:
        found = glob.glob(search_str)

    if replace_root:
        # may be necessary if not returning full path? -> TODO: add tests
        found = [f.replace(path, "root").split(os.path.sep) for f in found]

    if full_path:
        return found
    else:
        return [f[-1] for f in found]


def read_text(path: PathStr, file_format: str = "") -> dict:
    """Read a text file from specified path on either local fs or GCS."""
    if not file_format:
        file_format = Path(path).suffix
    read_func = _text_reader(file_format)
    if is_gcs(path):
        fs = FileClient.get_gcs_file_system()
        with fs.open(path, "r") as file:
            return read_func(file)
    else:
        with open(path) as file:
            return read_func(file)


def write_text(path: PathStr, content: str | dict, file_format: str) -> None:
    """Write json file to path on either local fs or GCS."""
    if not file_format:
        file_format = Path(path).suffix
    write = _text_writer(file_format)

    if is_gcs(path):
        fs = FileClient.get_gcs_file_system()
        with fs.open(path, "w") as file:
            write(file, content)
    else:
        mk_parent_dir(path)
        with open(path, "w") as file:
            write(file, content)


def _text_reader(file_format: str) -> Callable:
    """Helper function for selecting a function for reading files in the specified format."""
    match file_format.lower():
        case "json":

            def reader(file: PathStr) -> dict:
                return json.load(file)

        case "toml":

            def reader(file: PathStr) -> dict:
                return tomli.load(file)

    return reader


def _text_writer(file_format: str) -> Callable[[PathStr, dict], None]:
    """Return a function for writing a dict to text file in specified format.

    Args:
        file_format (str): Format of the text file, either "json" or "toml".

    Raises:
        ValueError: If the file format is not supported.
    """
    match file_format.lower():
        case "json":

            def writer(file: PathStr, content: dict | str) -> None:
                if isinstance(content, str):
                    content = json.loads(content)
                return json.dump(content, file, indent=4, ensure_ascii=False)

        case "toml":

            def writer(file: PathStr, content: dict | str) -> None:
                if isinstance(content, str):
                    content = tomli.loads(content)
                return tomli_w.dump(file, content)

        case _:
            raise ValueError(f"Format {file_format} is not supported.")

    return writer


def read_json(path: PathStr) -> dict:
    """Read json file from path on either local fs or GCS."""
    if is_gcs(path):
        fs = FileClient.get_gcs_file_system()
        with fs.open(path, "r") as file:
            return json.load(file)
    else:
        with open(path) as file:
            return json.load(file)


def write_json(path: PathStr, content: str | dict) -> None:
    """Write json file to path on either local fs or GCS."""
    if is_gcs(path):
        fs = FileClient.get_gcs_file_system()
        with fs.open(path, "w") as file:
            if isinstance(content, str):
                content = json.loads(content)
            json.dump(content, file, indent=4, ensure_ascii=False)
    else:
        mk_parent_dir(path)
        with open(path, "w") as file:
            json.dump(content, file, indent=4, ensure_ascii=False)


def read_parquet(
    path: PathStr,
    lazy: bool = False,  # TODO: Later steps need lazy-proofing before we can switch defaults.
    implementation: str = "pyarrow",
    **kwargs,
) -> narwhals.typing.Frame:
    """Read a Parquet file into a dataframe.

    This function can read from both local and GCS paths.

    Args:
        path: The path to the Parquet file.
        lazy: If True, returns a lazy dataframe. Defaults to False.
        implementation: The backend to use for reading the file.
            Defaults to "pyarrow".
        **kwargs: Additional keyword arguments passed to the backend.

    Returns:
        A Narwhals dataframe.
    """
    if lazy:
        return narwhals.scan_parquet(path, backend=implementation, **kwargs)
    else:
        return narwhals.read_parquet(path, backend=implementation, **kwargs)


def write_parquet(
    data: pyarrow.Table | IntoFrame,
    path: PathStr,
    schema: pyarrow.Schema | None = None,
    **kwargs,
) -> None:
    """Write a dataframe to a Parquet file.

    This function can write to both local and GCS paths, automatically
    selecting the correct filesystem backend. It also handles schema validation.

    Args:
        data: The dataframe to write (can be a PyArrow Table or any
            Narwhals-compatible dataframe).
        path: The destination path for the Parquet file.
        schema: An optional PyArrow schema to validate against before writing.
        **kwargs: Additional keyword arguments passed to the backend.
    """
    table = to_arrow(data, schema)  # to validate schema ...
    if is_gcs(path):
        fs = FileClient.get_gcs_file_system()
    else:
        fs = pyarrow.fs.LocalFileSystem()
        mk_parent_dir(path)

    if isinstance(table, pyarrow.Table):
        pq.write_table(
            table,
            where=path,
            filesystem=fs,
            # schema=schema,
            **kwargs,
        )
    else:
        # TODO: figure out how to do schema validation, then this would do:
        narwhals.from_native(data).write_parquet(path)
    # to make schema validation work / keep IO pure pyarrow it may bew better to go back to this(?):
    # pyarrow.dataset.write_dataset(
    #     data,
    #     path,
    #     filesystem=fs,
    #     format="parquet",
    #     schema=schema,
    #     # partitioning=["as_of_utc"],
    #     # partitioning_flavor="hive",
    #     **kwargs,
    # )
    # --> TODO: review interaction / lack thereof with pyarrow-...-helpers



"""The :py:mod:`ssb_timeseries` package is a helper library for production and analysis of statistical data in the form of *time series*.

It is designed to make it as easy as possible to store data and metadata for datasets and series in ways that are consistent with the :doc:`information model <../info-model>`, and to facilitate integration with automated workflows.

Functionality includes:

* Read and write data and metadata
* Metadata maintenance: tagging, detagging, retagging
* Search and filtering
* Time algebra: downsampling and upsampling to other time resolutions
* Linear algebra operations with sets (matrices) and series (column vectors)
* Metadata aware calculations, like unit conversions and aggregation over taxonomy hierarchies
* Basic plotting

The most practical entry points are the :py:mod:`ssb_timeseries.dataset` and :py:mod:`ssb_timeseries.catalog` modules.

.. automodule:: ssb_timeseries.dataset
    :noindex:
    :synopsis:
    :exclude-members: __init__

.. automodule:: ssb_timeseries.catalog
    :noindex:
    :synopsis:
    :exclude-members: __init__

The other modules of the package are helpers used by these core modules, and not intended for direct use.

Some notable exceptions are taxonomy and hierarchy features of :py:mod:`ssb_timeseries.meta` and type definitions in :py:mod:`ssb_timeseries.types`.
:py:mod:`ssb_timeseries.config` may be used for initial set up and later switching between repositories, if needed.
The :py:mod:`ssb_timeseries.io` seeks to make the storage agnostic of whether data and metada are stored in files or databases and :py:mod:`ssb_timeseries.fs` is an abstraction for local vs GCS file systems.
"""

from __future__ import annotations

from ssb_timeseries.catalog import get_catalog
from ssb_timeseries.config import Config
from ssb_timeseries.logging import set_up_logging_according_to_config

get_configuration = Config.active
"""Return the active configuration."""

logger = set_up_logging_according_to_config(__name__, get_configuration().logging)


__all__ = [
    "dataframes",
    "dataset",
    "dates",
    "get_catalog",
    "get_configuration",
    "io",
    "logger",
    "sample_data",
    "types",
]



"""A collection of helper functions to work with dataframes."""

from __future__ import annotations

from collections.abc import Iterable
from typing import Any
from typing import Literal
from typing import cast

import narwhals as nw
import polars as pl
import pyarrow
from narwhals.typing import Frame
from narwhals.typing import FrameT
from narwhals.typing import IntoFrame
from narwhals.typing import IntoFrameT
from numpy.typing import DTypeLike
from numpy.typing import NDArray

from .dates import standardize_dates
from .types import SeriesType
from .types import Temporality
from .types import Versioning

# mypy: disable-error-code="union-attr"


def _coalesce(*args):
    """Return first value that evaluates to True."""
    return next((bool(arg) for arg in args if arg is not None), None)


def copy(df: FrameT) -> Any:
    """Return an (eager) copy of the dataframe."""
    return eager(df).clone().to_native()


def eager(df: Frame) -> Any:
    """Collect or compute for lazy implementations."""
    return nw.from_native(df).lazy().collect()


def empty_frame(
    *,
    columns: list[str] | None = None,
    schema: pyarrow.Schema | None = None,
    implementation: str = "arrow",
) -> Any:
    """Return a dataframe or Arrow table with no data."""
    if schema:
        pa_tbl = pyarrow.Table.from_pylist([], schema=schema)
        df = nw.from_native(pa_tbl)
    elif columns:
        # Pandas dependency to be removed,
        # but Pyarrow implementsation below does not create columns :(
        import pandas

        pd_df = pandas.DataFrame(columns=columns)
        df = nw.from_native(pd_df)  # type: ignore
    elif columns and False:
        pa_schema = pyarrow.schema([])
        for c in columns:
            if c in ["as_of", "valid_at", "valid_from", "valid_to"]:
                pa_schema.append(pyarrow.field(c, pyarrow.date64()))
            else:
                pa_schema.append(pyarrow.field(c, pyarrow.float64()))
        pa_tbl = pyarrow.Table.from_pylist([], schema=pa_schema)
        df = nw.from_native(pa_tbl)
    else:
        df = nw.from_native(pyarrow.Table.from_pylist([]))

    match implementation.lower():
        case "arrow" | "pyarrow" | "pa":
            return df.to_arrow()
        case "pandas" | "pd":
            return df.to_pandas()
        case "polars" | "pl":
            return df.to_polars()


def are_equal(*frames: IntoFrame) -> bool:
    """Check if dataframes are equal."""
    first_df = nw.from_native(frames[0]).to_polars()
    for df in frames[1:]:
        df = nw.from_native(df).to_polars()
        if df.shape != first_df.shape:
            return False
        elif set(df.columns) != set(first_df.columns):
            return False
        elif not df.select(sorted(df.columns)).equals(
            first_df.select(sorted(first_df.columns))
        ):
            return False
    return True


def infer_datatype(df: IntoFrame, **kwargs) -> SeriesType:
    """Checks dataframe columns and kwargs to identify SeriesType.

    Args:
        df: The dataframe to check.
        **kwargs: Override values for keys ' versioning', 'temporality'.

    Returns:
        SeriesType if the dataframe and/or kwargs provides sufficient information to infer Versioning and Temporality.

    `Versioning` is determined by assessing in order of (priority):
    - the kwarg value for the key 'versioning', if provided.
    - if dataframe columns contain 'as_of', 'as_of_tz' or 'as_of_utc'.
    - if keys of kwargs contain 'as_of', 'as_of_tz' or 'as_of_utc'.

    `Temporality` is determined by assessing in order of (priority):
    - the kwarg value for the key 'temporality', if provided.
    - if the dataframe columns contain 'valid_from' and 'valid_to'.

    """
    nw_df = nw.from_native(df)
    vs_markers = {"as_of", "as_of_tz", "as_of_utc"}
    from_to = {"valid_from", "valid_to"}

    columns = set(nw_df.columns)

    vs_explicit = kwargs.get("versioning")
    vs_from_kwargs = Versioning.AS_OF if vs_markers & set(kwargs.keys()) else None
    vs_from_column = Versioning.AS_OF if vs_markers & columns else None
    versioning = _coalesce(vs_explicit, vs_from_column, vs_from_kwargs, Versioning.NONE)

    t_from_columns = (
        Temporality.FROM_TO if from_to & columns == from_to else Temporality.AT
    )
    temporality = _coalesce(kwargs.get("temporality"), t_from_columns)

    return SeriesType(versioning, temporality)


def is_df_like(obj: Any) -> bool:
    """Checks if an object is "dataframe-like" for Narwhals compatibility.

    This is a robust, duck-typing alternative to `isinstance(obj, IntoFrameT)`,
    which is not possible.

    Args:
        obj: The object to check.

    Returns:
        True if the object has dataframe-like attributes, False otherwise.
    """
    # All supported dataframe objects have a `.shape` tuple (rows, cols)
    # and a `.columns` attribute (a list or index of column names).
    # We also check that the object is not a NumPy array, as arrays also
    # have a .shape but are not dataframes.
    return (
        hasattr(obj, "shape")
        and isinstance(getattr(obj, "shape", None), tuple)
        and hasattr(obj, "columns")
        and "numpy.ndarray" not in str(type(obj))  # A simple way to exclude numpy
    )


def is_empty(df: IntoFrame) -> bool:
    """Check if dataframe is empty."""
    # nox/mypy vs 1.10.1 --> [redundant-cast] | pre-commit --> [no-any-return]
    # (but cast was introduced because of other error with other mypy env)
    # return cast(bool, nw.from_native(df).is_empty())  # nox --> [redundant-cast]
    # return nw.from_native(df).is_empty() # pre-commit --> [no-any-return]
    # ... fix(?):
    nw_df = nw.from_native(df)
    df_is_empty = eager(nw_df).is_empty()
    return cast(bool, df_is_empty)


def rename_columns(
    df: Any,
    substitutions: dict[str, str],
) -> Any:
    """Rename columns of dataframe."""
    nw_df = nw.from_native(df)
    names = nw_df.schema.names()

    mapping = {}
    for n in names:
        for r, w in substitutions.items():
            if r in n:
                mapping[n] = n.replace(r, w)

    return nw_df.rename(mapping).to_native()


def to_arrow(
    df: IntoFrame,
    schema: pyarrow.Schema | None = None,
) -> pyarrow.Table:
    """Convert any Narwhals compatible Data Frame to Pyarrow table, cast schema if provided."""
    table = nw.from_native(df).to_arrow()
    if schema:
        return table.select(schema.names).cast(schema)
    else:
        return table


def to_numpy(
    df: Frame,
    dtype: DTypeLike | None = None,
    *,
    output_type: Literal["homogeneous", "structured"] = "homogeneous",
) -> NDArray:
    """Converts a dataframe to a NumPy ndarray.

    Parameters:
        dtype : type | DTypeLike | None
            The desired dtype for the resulting array. If None and output_type is
            'homogeneous', will upcast to 'object' if columns have mixed types.
            dtype is ignored when output_type='structured' as column dtypes are preserved.
        output_type : "homogeneous" | "structured"
            'homogeneous': Returns a 2D array, possibly with dtype='object' for mixed types.
            'structured': Returns a 1D structured array, preserving individual column dtypes.

    Returns:
        numpy.ndarray
            A NumPy array representation of the data.
    """
    nw_df = nw.from_native(df)
    if output_type == "homogeneous":
        return cast(NDArray, nw_df.to_arrow().to_numpy(dtype=dtype, copy=True))
    elif output_type == "structured":
        return nw_df.to_pandas().to_records(index=False)
    else:
        raise ValueError(
            f"Invalid output_type: {output_type}. Must be 'homogeneous' or 'structured'."
        )


def merge_data(
    old: IntoFrameT,
    new: IntoFrameT,
    date_cols: Iterable[str],
    **kwargs,  # temporality: Temporality = Temporality.AT,
) -> pyarrow.Table:
    """Merge new data into an existing dataframe, handling overlaps for period-based data.

    For `AT` temporality, it keeps the last entry for duplicates based on date columns.
    For `FROM_TO` temporality, it uses an anti-join to replace rows with matching
    `valid_from` and `valid_to` pairs.
    """
    new_utc = standardize_dates(new)
    old_utc = standardize_dates(old)
    new_pl = cast(nw.DataFrame, nw.from_native(new_utc)).to_polars()
    old_pl = cast(nw.DataFrame, nw.from_native(old_utc)).to_polars()
    common_date_cols = [
        col for col in date_cols if col in old_pl.columns and col in new_pl.columns
    ]
    if not common_date_cols:
        raise ValueError(
            f"No matching date columns; old:\n{old_pl.schema}\nnew:\n{new_pl.schema}."
        )

    new_pl = new_pl.with_columns(pl.col(pl.Float32).cast(pl.Float64))
    old_pl = old_pl.with_columns(pl.col(pl.Float32).cast(pl.Float64))

    old_filtered = old_pl.join(new_pl, on=common_date_cols, how="anti")

    merged = nw.concat(
        [nw.from_native(old_filtered), nw.from_native(new_pl)],
        how="diagonal",
        # )
        # out = merged.unique(
        #    subset=common_date_cols,
        #    keep="last",
    ).sort(by=sorted(common_date_cols))
    pa_table = merged.to_arrow()
    return cast(pyarrow.Table, pa_table)



"""The :py:mod:`ssb_timeseries.dataset` module and its :py:class:`Dataset` class is the very core of the :py:mod:`ssb_timeseries` package, defining most of the key functionality.

The dataset is the unit of analysis for both :doc:`information model <../info-model>` and :doc:`workflow integration <../workflow>`,and performance will benefit from linear algebra with sets as matrices consisting of series column vectors.

As described in the :doc:`../info-model` time series datasets may consist of any number of series of the same :py:class:`~ssb_timeseries.types.SeriesType`.
The series types are defined by dimensionality characteristics:

* :py:class:`~ssb_timeseries.types.Versioning` (NONE, AS_OF, NAMED)
* :py:class:`~ssb_timeseries.types.Temporality` (Valid AT point in time, or FROM and TO for duration)
* The type of the value. For now only scalar values are supported.

Additional type determinants (sparsity, irregular frequencies, non-numeric or non-scalar values, ...) are conceivable and may be introduced later.
The types are crucial because they are reflected in the physical storage structure.
That in turn has practical implications for how the series can be interacted with, and for methods working on the data.

.. admonition:: See also:
    :class: more

    The :py:mod:`ssb_timeseries.catalog` module for tools for searching for datasets or series by names or metadata.
"""

from __future__ import annotations

import re
import warnings
from collections.abc import Iterable
from collections.abc import Sequence
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from typing import cast
from typing import no_type_check

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self  # noqa: UP035 #backport to 3.10

import matplotlib.pyplot as plt  # noqa: F401
import narwhals as nw
import narwhals.selectors as ncs
import numpy as np
import pyarrow as pa
from narwhals.typing import Frame
from narwhals.typing import FrameT
from narwhals.typing import IntoDType
from narwhals.typing import IntoFrame
from narwhals.typing import IntoFrameT
from narwhals.typing import IntoSeries
from numpy.typing import DTypeLike
from numpy.typing import NDArray

from ssb_timeseries.types import SeriesType

from . import io
from . import meta
from .config import Config
from .dataframes import empty_frame
from .dataframes import infer_datatype
from .dataframes import is_df_like
from .dataframes import is_empty
from .dataframes import rename_columns
from .dates import date_local
from .dates import date_utc
from .dates import period_index
from .dates import utc_iso
from .logging import logger
from .types import F
from .types import PathStr
from .types import Temporality
from .types import Versioning

if TYPE_CHECKING:
    import pandas as pd
    import polars as pl

# mypy: disable-error-code="assignment,attr-defined,union-attr,arg-type,call-overload,no-untyped-call,dict-item,no-untyped-def,no-any-return"
# ruff: noqa: RUF013


def default_repository() -> str:
    """Check the configuration and get the name of the default repository."""
    repos = Config.active().repositories
    candidates = []
    for name, cfg in repos.items():
        if cfg.get("default", False):
            return str(name)
        elif cfg.get("directory") and cfg.get("catalog"):
            candidates.append(str(name))

    if len(candidates) == 1:
        return candidates[0]
    elif len(candidates) == 0:
        raise LookupError("Default repository could not be identified.")
    elif len(candidates) > 1:
        warnings.warn(
            f"Repositories {candidates} all specify data and metadata handling, but none of them is marked as default. The first on the list is used.",
            stacklevel=2,
        )
    return candidates[0]


def _identify_data_type(**kwargs: Any) -> SeriesType | None:
    """Analyse kwargs to determine if data type is supplied or can be inferred."""
    data_type = kwargs.get("data_type", "")
    data = kwargs.get("data")
    existing_tags = kwargs.pop("existing", {})

    if existing_tags:
        existing = SeriesType(
            existing_tags.get("versioning"), existing_tags.get("temporality")
        )
    else:
        existing = None

    if data_type and isinstance(data_type, SeriesType):
        provided = data_type
    elif data_type:
        provided = SeriesType(data_type)
    elif data is not None:
        provided = infer_datatype(data, **kwargs)
    else:
        provided = None

    def coalesce(*args):
        return next((arg for arg in args if arg is not None), None)

    return coalesce(provided, existing)


class Dataset:
    """Datasets contain one or more series of the same :py:class:`~ssb_timeseries.types.SeriesType`.

    The grouping of series in a set usually reflect some common denominator in terms of descriptive metadata.
    In a well defined set, all the series in the set come from the same process,
    so that the set comprises a natural chunk of data for reads and writes in a batch oriented workflow.

    For all the series in a dataset to be of the same :py:class:`~ssb_timeseries.types.SeriesType` means they share dimensionality characteristics :py:class:`~ssb_timeseries.types.Versioning` and :py:class:`~ssb_timeseries.types.Temporality` and any other schema information that have technical implications.
    See the :doc:`../info-model` documentation for more about that.

    Descriptive commonality is beneficial,
    especially for attributes that reflect technical types,
    although not technically enforced.
    A notable example is time resolution.
    Sparse data is a strong indication that a dataset is not well defined.
    Uniform resolution help minimize the number of gaps in the series.
    'Gaps' in this context is any representation of undefined values:
    None, null, NAN or "not a number" values, as opposed to the number zero.
    (The number zero is a gray area - it can be perfectly valid, but can also be an indication that not all the series should be part of the same set.)

    :var str name: The name of the set.
    :var SeriesType data_type: The type of the contents of the set.
    :var datetime as_of_tz: The version datetime, if applicable to the ``data_type``.
    :var Dataframe data: A dataframe or table structure with one or more datetime columns defined by ``datatype`` and a column per series in the set.
    :var dict tags: A dictionary with metadata describing both the dataset itself and the series in the set.

    .. admonition:: Maintaining tags
        :class: more dropdown

        |tagging|

    """

    name: str
    as_of_utc: datetime | None
    data_type: SeriesType
    data: Any
    tags: dict
    repository: str
    sharing: dict | None
    lineage: str | None

    def __init__(
        self,
        name: str,
        as_of_tz: datetime | None = None,
        *,
        data_type: SeriesType | Sequence[str] | None = None,
        repository: str = "",
        find_existing: bool = True,
        **kwargs: Any,
    ) -> None:
        """Retrieve an existing dataset or create a new one.

        To create a new set, data_type must be specified,
        and if data_type versioning is as AS_OF, also a datetime.
        Providing an AS_OF date has no effect if versioning is NONE.
        If not, but data is passed, :py:meth:`as_of_tz` defaults to current time.
        All dates should be timezone aware.
        If no timezone is provided, a default is assumed (CET).

        The `data` parameter accepts a dataframe with one or more date columns and one column per series.
        Initially only Pandas was supported, but this dependency is about to be relaxed to include other implementations of the samedata structure.
        Beyond Polars and Pyarrow, notable options under consideration include Ibis, DuckDB and Narwhals.

        Data is kept in memory and not stored before explicit call to :py:meth:`save`.
        The data is stored in parquet files, with JSON formatted metadata in the header.

        Metadata will always be read if the set exists.
        When loading existing sets, find_existing = False will suppress reading large amounts of data.
        For data_types with AS_OF versioning, not providing the AS_OF date will have the same effect.

        If series names can be mapped to metadata, the keyword arguments ``attributes``, ``separator`` and ``regex`` will, if provided, be passed through to :py:class:`~Dataset.series_names_to_tags`.
        If series names are not easily translated to tags, :py:class:`~Dataset.tag_dataset` and :py:class:`~Dataset.tag_series` and their siblings :py:meth:`retag <Dataset.retag_series>` and :py:meth:`detag <Dataset.detag_series>` can be used for manual meta data maintenance.

        :keyword list[str] attributes: Attribute names for use with :py:class:`~Dataset.series_names_to_tags` in combination with either ``separator`` or ``regex``.
        :keyword str separator: Character(s) separating ``attributes`` for use with :py:class:`~Dataset.series_names_to_tags`.
        :keyword str regex: Regular expression with capture groups corresponding to ``attributes``. Used instead of the separator to match more complicated name patterns in :py:class:`~Dataset.series_names_to_tags`.

        .. admonition:: Maintaining tags
           :class: more dropdown

           |tagging|


        .. code::

            import ssb_timeseries as ts
            df = ts.sample_data.xyz_at()
            print(df)

            x = ts.dataset.Dataset(
                name='mydataset',
                data_type=SeriesType.simple(),
                data=df
            )

        .. testoutput::
            :hide:

            34

        """
        if find_existing:
            tags_for_existing = io.find(
                set_name=name,
                repository=repository,
            )
            if tags_for_existing is None:
                tags_for_existing = {}
            elif isinstance(tags_for_existing, dict):
                type_of_existing = SeriesType(
                    versioning=tags_for_existing.get("versioning"),
                    temporality=tags_for_existing.get("temporality"),
                )
            elif isinstance(tags_for_existing, list):
                raise LookupError(
                    f"Dataset.__init__ can not get existing metadata because search returned multiple sets of same name: {name=}:\n{tags_for_existing}."
                )
        else:
            tags_for_existing = {}

        data_type = _identify_data_type(
            **{
                **kwargs,
                "data_type": data_type,
                "as_of_tz": as_of_tz,
                "existing": tags_for_existing,
            }
        )
        logger.debug(
            "DATASET.debug found existing dataset: %s of type %s in repository %s",
            tags_for_existing.get("name"),
            (tags_for_existing.get("versioning"), tags_for_existing.get("temporality")),
            tags_for_existing.get("repository"),
        )

        self.name = name
        self.data_type = data_type
        self.tags = {}

        self.repository = next(
            (
                r
                for r in [
                    repository,
                    tags_for_existing.get("repository", False),
                ]
                if r
            ),
            default_repository(),
        )
        if tags_for_existing:
            self.tags = tags_for_existing

        if data_type:
            self.data_type = data_type

        if as_of_tz is not None:
            self.as_of_utc = date_utc(as_of_tz)
        else:
            self.as_of_utc = None

        identify_latest: bool = (
            self.data_type.versioning == Versioning.AS_OF
            and not as_of_tz
            and find_existing
        )
        if identify_latest:
            logger.debug(
                "Init %s '%s' without 'as_of_tz' --> identifying latest.",
                self.data_type,
                self.name,
            )
            lookup_as_of = self.versions()[-1]
            if isinstance(lookup_as_of, datetime):
                as_of_tz = lookup_as_of
            self.as_of_utc = date_utc(as_of_tz)
        elif as_of_tz:
            self.as_of_utc = date_utc(as_of_tz)
        else:
            self.as_of_utc = None

        kwarg_data = kwargs.get("data", None)
        if is_df_like(kwarg_data) and not is_empty(kwarg_data):
            self.data = kwarg_data
        elif find_existing:  # and self.data_type.versioning == types.Versioning.AS_OF:
            self.data = io.DataIO(self).dh.read()
        else:
            self.data = empty_frame()

        stored_tags = tags_for_existing
        self.tags = self.default_tags()
        self.tags.update(stored_tags)
        self.tag_dataset(tags=kwargs.get("dataset_tags", {}))

        # all of the following should be turned into set level tags - or find another way into the parquet files?

        # autotag:
        attributes = kwargs.get("attributes", [])
        separator = kwargs.get("separator", "_")
        regex = kwargs.get("regex", "")
        ready_to_auto_tag = _has_auto_tag_information(self, locals())
        apply_to_all = kwargs.get("series_tags", {})
        self.auto_tag_config = {
            "attributes": attributes,
            "separator": separator,
            "regex": regex,
        }
        if not is_empty(self.data):
            self.tag_series(tags=apply_to_all)
            if ready_to_auto_tag:
                self.series_names_to_tags()

        # "owner" / sharing / access
        self.product: str = kwargs.get("product", "")
        self.process_stage: str = kwargs.get("process_stage", "")
        self.sharing: dict[str, str] = kwargs.get("sharing", {})

    def copy(
        self,
        new_name: str = "",
        select: str | list[str] = "",
        **kwargs: Any,
    ) -> Self:
        """Create a copy of the Dataset.

        The copy will be created with the same data_type, as_of_tz, data, and tags.
        The name will be altered as a safeguard against accidentally overwriting existing data.
        Tags are updated accordingly, using 'Dataset.rename()'.

        Args:
            new_name: An optional new name for the copied dataset.
            select: An optional list of columns to include in the copy.
            **kwargs: Additional keyword arguments passed to underlying functions.

        Keyword Args:
            allow providing parameters for initializing the copied set.
        """
        if not new_name:
            new_name = f"COPY of {self.name}"

        data = kwargs.pop("data", self.data)
        match select.lower():
            case "dates_only":
                raise NotImplementedError("TODO!")
            case "select":
                raise NotImplementedError("TODO!")
            case "all" | _:
                ...  # data = data

        # Observed here: OSError: [Errno 36] File name too long
        # New names from .select() or column aggregations easily become too long.
        # A workaround was applied in fs.exists()
        # out = deepcopy(self)
        # out.rename(new_name)
        autotag_attr = kwargs.pop("attributes", self.auto_tag_config["attributes"])
        out = self.__class__(
            name=new_name,
            data_type=kwargs.pop("data_type", self.data_type),
            data=data,
            find_existing=False,
            attributes=autotag_attr,
            **kwargs,
        )
        # not necessary?
        # for k, v in self.__dict__.items():
        #    #print(f"copy attr from self: {k=}, ")
        #    setattr(out, k, v)

        # lineage coalesce: kwarg, self, new_name
        lineage = getattr(self, "lineage", new_name)
        out.lineage = kwargs.pop("lineage", lineage)

        copied_series_tags = out.tags["series"]
        tags_to_eliminate = set(copied_series_tags.keys()) - set(out.series)
        for series_key in tags_to_eliminate:
            copied_series_tags.pop(series_key)
        out.tags["series"] = copied_series_tags
        return out

    def rename(
        self,
        /,
        new_set_name: str = "",
        *series_name_substitutions: tuple[str, str],
    ) -> None:
        """Rename the Dataset.

        For use by .copy, and on very rare other occasions. Does not move or rename any previously stored data.
        """
        if new_set_name:
            self.name = new_set_name

            self.tags["name"] = new_set_name
            for _, v in self.tags["series"].items():
                v["dataset"] = new_set_name

        for subst in series_name_substitutions:
            new_series_tags = {}
            for old_name, tags in self.tags["series"].items():
                new_name = old_name.replace(subst[0], subst[1])
                tags["name"] = new_name
                new_series_tags[new_name] = tags
            self.tags["series"] = new_series_tags
            self.data = rename_columns(
                self.data,
                {subst[0]: subst[1]},
            )

    def save(self, as_of_tz: datetime = None) -> None:
        """Persist the Dataset.

        Args:
            as_of_tz (datetime): Provide a timezone sensitive as_of date in order to create another version. The default is None, which will save with Dataset.as_of._utc (utc dates under the hood).
        """
        if as_of_tz is not None:
            self.as_of_utc = date_utc(as_of_tz)

        io.save(self)

    def snapshot(self) -> None:
        """Copy data snapshot to immutable processing stage bucket and shared buckets.

        If :py:attr:`~ssb_timeseries.dataset.Dataset.sharing` identifies a configured location,
        the snapshot files will be copied there.

        See :py:func:`~ssb_timeseries.io.persist` for more detail.
        """
        io.persist(self)  # is 'archive' a better name than 'persist' or 'snapshot'?

    def versions(self, **kwargs: Any) -> list[datetime | str]:
        """Get list of all series version markers (`as_of` dates or version names).

        By default `as_of` dates will be returned in local timezone. Provide `return_type = 'utc'` to return in UTC, 'raw' to return as-is.
        """
        versions = io.versions(self)
        if not versions:
            return []
        else:
            logger.debug("DATASET %s: versions: %s.", self.name, versions)

        if self.data_type.versioning == Versioning.AS_OF:
            return_type = kwargs.get("return_type", "local")
        else:
            return_type = "raw"

        match return_type:
            case "local":
                return [date_local(v) for v in versions]
            case "utc":
                return [date_utc(v) for v in versions]
            case "utc_iso":
                return [utc_iso(v) for v in versions]
            case _:
                return versions

    @property
    def series(self) -> list[str]:
        """Get (sorted) series names, ie. all columns that are not time related.

        As of version 0.6 this implies numeric, with a notable exception for booleans.
        While booleans occur naturally inside calculations and some logical functions have been added,
        the support for booleans is not complete.

        Other datatypes are under consideration, but even beyond numeric and boolean values,
        it is likely to remain a requirement that datasets remain a single type.
        Ie that `series` yields the same result as one of the specialized functions `numeric_columns` or `boolean_columns`.
        """
        if not is_empty(self.data):
            dt_expr = ~ncs.by_dtype(nw.Datetime, nw.Date)
            non_datetime_cols = self.nw.select(dt_expr).columns
            return sorted(non_datetime_cols)
        elif self.tags and "series" in self.tags:
            return sorted(self.tags["series"].keys())
        return []

    @property
    def series_tags(self) -> meta.SeriesTagDict:
        """Get series tags."""
        return self.tags["series"]  # type: ignore

    def default_tags(self) -> meta.DatasetTagDict:
        """Return default tags for set and series."""
        return {
            "name": self.name,
            "versioning": str(self.data_type.versioning),
            "temporality": str(self.data_type.temporality),
            "series": {s: {"dataset": self.name, "name": s} for s in self.series},
            "repository": self.repository,
        }

    def tag_dataset(
        self,
        tags: meta.TagDict = None,
        **kwargs: str | list[str] | set[str],
    ) -> None:
        """Tag the set.

        Tags may be provided as dictionary of tags, or as kwargs.

        In both cases they take the form of attribute-value pairs.

        Attribute (str): Attribute identifier.
        Ideally attributes relies on KLASS, ie a KLASS taxonomy defines the possible attribute values.

        Value (str): Element identifier, unique within the taxonomy. Ideally KLASS code.

        Note that while no such restrictions are enforced, it is strongly recommended that both attribute names (``keys``) and ``values`` are standardised.
        The best way to ensure that is to use taxonomies (for SSB: KLASS code lists).
        However, custom controlled vocabularies can also be maintained in files.

        .. admonition:: Maintaining tags
            :class: more dropdown

            |tagging|


        Examples:
            >>> from ssb_timeseries.dataset import Dataset
            >>> from ssb_timeseries.types import SeriesType
            >>> from ssb_timeseries.sample_data import create_df
            >>>
            >>> x = Dataset(name='sample_dataset',
            >>>         data_type=SeriesType.simple(),
            >>>         data=create_df(['x','y','z'],
            >>>             start_date='2024-01-01',
            >>>             end_date='2024-12-31',
            >>>             freq='MS',)
            >>> )
            >>>
            >>> x.tag_dataset(tags={'country': 'Norway', 'about': 'something_important'})
            >>> x.tag_dataset(another_attribute='another_value')


        """
        if not self.__getattribute__("tags"):
            # should not be possible, hence
            raise ValueError(f"Tags not defined for dataset: {self.name}.")

        if not tags and not kwargs:
            return
        elif not tags:
            tags = {}
        if kwargs:
            tags.update(**kwargs)

        propagate = tags.pop("propagate", True)
        if tags:
            self.tags = meta.add_tag_values(self.tags, tags, recursive=propagate)

    def tag_series(
        self,
        names: str | list[str] = "*",
        tags: meta.TagDict = None,
        **kwargs: str | list[str],
    ) -> None:
        """Tag the series identified by ``names`` with provided tags.

        Tags may be provided as dictionary of tags, or as kwargs.

        In both cases they take the form of attribute-value pairs.

        Attribute (str): Attribute identifier.
        Ideally attributes relies on KLASS, ie a KLASS taxonomy defines the possible attribute values.

        Value (str): Element identifier, unique within the taxonomy. Ideally KLASS code.

        If series names follow the same pattern of attribute values in the same order separated by the same character sequence, tags can be propagated accordingly by specifying ``attributes`` and ``separator`` parameters. The separator will default to underscore if not provided. Note that propagation by pattern will affect *all* series in the set, not only the ones identified by ``names``.

        .. admonition:: Maintaining tags
            :class: more dropdown

            |tagging|

        Examples:
            Dependencies

            >>> from ssb_timeseries.dataset import Dataset
            >>> from ssb_timeseries.types import SeriesType
            >>> from ssb_timeseries.sample_data import create_df
            >>>
            >>> some_data = create_df(['x', 'y', 'z'], start_date='2024-01-01', end_date='2024-12-31', freq='MS')

            Tag by kwargs

            >>> x = Dataset(name='sample_set',data_type=SeriesType.simple(),data=some_data)
            >>> x.tag_series(example_1='string_1', example_2=['a', 'b', 'c'])

            Tag by dict

            >>> x = Dataset(name='sample_set',data_type=SeriesType.simple(),data=some_data)
            >>> x.tag_series(tags={'example_1': 'string_1', 'example_2': ['a', 'b', 'c']})

        """
        if not tags and not kwargs:
            return

        if not tags:
            tags = {}
        tags.update(kwargs)

        if names == "*":
            names = self.series
        elif isinstance(names, str):
            names = [names]

        inherit_from_set_tags = meta.inherit_set_tags(self.tags)

        for n in names:
            if not self.tags["series"][n]:
                self.tags["series"][n] = {"name": n}
            self.tags["series"][n].update({**inherit_from_set_tags, **tags})

    def detag_dataset(
        self,
        *args: str,
        **kwargs: Any,
    ) -> None:
        """Detag selected attributes of the set.

        Tags to be removed may be provided as list of attribute names or as kwargs with attribute-value pairs.

        .. admonition:: Maintaining tags
            :class: more dropdown

            |tagging|
        """
        self.tags = meta.delete_dataset_tags(
            self.tags,
            *args,  # [a for a in args if isinstance(a, str)]
            propagate=True,
            **kwargs,
        )

    @no_type_check  # "operator"
    def detag_series(
        self,
        *args: str,
        **kwargs: Any,
    ) -> None:
        """Detag selected attributes of series in the set.

        Tags to be removed may be specified by args or kwargs.
        Attributes listed in `args` will be removed from all series.

        For kwargs, attributes will be removed from the series if the value matches exactly. If the value is a list, the matching value is removed. If kwargs contain all=True, all attributes except defaults are removed.

        .. admonition:: Maintaining tags
            :class: more dropdown

            |tagging|
        """
        self.tags["series"] = meta.delete_series_tags(
            self.tags["series"],
            *args,
            **kwargs,
        )

    @no_type_check
    def series_names_to_tags(
        self,
        attributes: list[str] | None = None,  # /NOSONAR
        separator: str = "",  # /NOSONAR
        regex: str = "",  # /NOSONAR
    ) -> None:
        """Tag all series in the dataset based on a series 'attributes', ie a list of attributes matching positions in the series names when split on 'separator'.

        Alternatively, a regular expression with groups that match the attributes may be provided.
        Ideally attributes relies on KLASS, ie a KLASS taxonomy defines the possible attribute values.

        Value (str): Element identifier, unique within the taxonomy. Ideally KLASS code.

        .. admonition:: Maintaining tags
            :class: more dropdown

            |tagging|

        Examples:
            >>> from ssb_timeseries.dataset import Dataset
            >>> from ssb_timeseries.types import SeriesType
            >>> from ssb_timeseries.sample_data import create_df

            Tag using attributes and dcefault separator:

            Let us create some data where the series names are formed by the values ['x', 'y', 'z']
            separated from ['a', 'b', 'c'] by an underscore:

            >>> some_data = create_df(
            >>>     ["x_a", "y_b", "z_c"],
            >>>     start_date="2024-01-01",
            >>>     end_date="2024-12-31",
            >>>     freq="MS",
            >>> )

            Then put it into a dataset and tag:

            >>> p = Dataset(
            >>>     name="sample_set",
            >>>     data_type=SeriesType.simple(),
            >>>     data=some_data,
            >>> )
            >>> p.series_names_to_tags(attributes=['XYZ', 'ABC'])

            >>> p.tags

            The above approach may be used at any time to add tags for an existing dataset, but the same arguments can also be provided when initialising the set:

            >>> z = Dataset(
            >>>     name="copy_of_sample_set",
            >>>     data_type=SeriesType.simple(),
            >>>     data=some_data,
            >>>     attributes=['XYZ', 'ABC'],
            >>> )

            Best practice is to do this only in the process that writes data to the set.
            For a finite number of series, it does not need to be repeated.

            If, on the other hand, the number of series can change over time, doing so at the time of writing ensures all series are tagged.
            Tag using attributes and regex:

            If series names are less well formed, a regular expression with groups matching the attribute list can be provided instead of the separator parameter.

            >>> more_data = create_df(
            >>>     ["x_1,,a", "y...b..", "z..1.1-23..c"],
            >>>     start_date="2024-01-01",
            >>>     end_date="2024-12-31",
            >>>     freq="MS",
            >>> )
            >>> x = Dataset(
            >>>     name="bigger_sample_set",
            >>>     data_type=SeriesType.simple(),
            >>>     data=more_data,
            >>> )
            >>> x.series_names_to_tags(attributes=['XYZ', 'ABC'], regex=r'([a-z])*([a-z])')

        """
        if attributes:
            self.auto_tag_config["attributes"] = attributes
        else:
            attributes = self.auto_tag_config["attributes"]

        if separator:
            self.auto_tag_config["separator"] = separator
        else:
            separator = self.auto_tag_config["separator"]

        if regex:
            self.auto_tag_config["regex"] = regex
        else:
            regex = self.auto_tag_config["regex"]

        apply_all = self.auto_tag_config.get("apply_to_all", {})
        inherited = meta.inherit_set_tags(self.tags)
        self.tag_series({**inherited, **apply_all})

        # always apply tags inherited from dataset and tags specificly applied to all series first
        for series_key in self.series:
            self.tags["series"].get(series_key, {}).update(inherited)
            if apply_all:
                self.tags["series"].get(series_key, {}).update(apply_all)

        if attributes:
            for series_key in self.tags["series"].keys():
                if regex:
                    name_parts = re.search(regex, series_key).groups()
                elif separator:
                    name_parts = series_key.split(separator)
                else:
                    raise AttributeError(
                        "DATASET.series_names_to_tags() requires either a regex or a separator."
                    )

                for attribute, value in zip(attributes, name_parts, strict=False):
                    # not necessary? self.tags["series"][series_key][attribute] = deepcopy(value)
                    self.tags["series"][series_key][attribute] = value
        else:
            logger.warning(
                "DATASET.series_names_to_tags() requires attributes to be defined for the Dataset object or to be passed as an argument."
            )
            # raise AttributeError( "Attributes must be defined in Dataset.auto_tag_config or passed as an argument.")

    def replace_tags(
        self,
        *args: tuple[meta.TagDict, meta.TagDict],
    ) -> None:
        """Retag selected attributes of series in the set.

        The tags to be replaced and their replacements should be specified in tuple(s) of :py:type:`tag dictionaries <ssb_timeseries.meta.TagDict>`;
        each argument in ``*args`` should be on the form ``({<old_tags>},{<new_tags>})``.

        Both old and new :py:type:`TagDict` can contain multiple tags.
         - Each tuple is evaluated independently for each series in the set.
         - If the tag dict to be replaced contains multiple tags, all must match for tags to be replaced.
         - If the new tag dict contains multiple tags, all are added where there is a match.

        .. admonition:: Maintaining tags
            :class: more dropdown

            |tagging|
        """
        for a in args:
            old = a[0]
            new = a[1]

            self.tags = meta.replace_dataset_tags(self.tags, old, new, recursive=True)

    @no_type_check
    def select(
        self,
        *names: str | list[str],
        pattern: str = "",
        regex: str = "",
        tags: (meta.TagDict | list[meta.TagDict]) = None,
        output: str = "dataset",
        new_name: str = "",
        **kwargs: Any,
    ) -> IntoFrameT | Self:
        """Select series (columns) of Dataset.data by textual pattern, regex or metadata tag dictionary.

        Or a combination.

        Args:
            names (str|list[str]): Series names

        Keyword Args:
            pattern: Text pattern for search 'like' in column names. Defaults to ''.
            regex: Expression for regex search in column names. Defaults to ''.
            tags: Dictionary with tags to search for. Defaults to None. All tags in dict must be satisfied for the same series (tags are combined by AND). If a list of values is provided for a tag, the criteria is satisfied for either of them (OR).
                | list(dict) Support for list(dict) is planned, not yet implemented, to satisfy alternative sets of criteria (the dicts will be combined by OR).
            output: Output type - dataset or dataframe.(df). Defaults to 'dataset'. Short forms 'df' or 'ds' are accepted.
            new_name: Name of new Dataset. If not provided, a new name is generated.
            **kwargs: if provided, goes into the init of the new set.

        Returns:
            Dataset | Dataframe:
            By default a new Dataset (a deep copy of self).
            If output="dataframe" or "df", a dataframe.
            TODO: Explore shallow copy / nocopy options.
        """
        if not any([names, pattern, regex, tags]):
            error_message = f"DATASET.select() was called without valid criteria:\n{names=}, {pattern=}, {regex=}, {tags=}"
            raise ValueError(error_message)

        expressions = [nw.col(self.datetime_columns)]
        if names:
            expressions.append(nw.col(*names))

        if regex:
            expressions.append(ncs.matches(regex))

        if pattern:
            expressions.append(ncs.matches(f".*{pattern}.*"))

        if tags:
            if isinstance(tags, list):
                matching_series = meta.search_by_tags(self.tags["series"], *tags)
            else:
                matching_series = meta.search_by_tags(self.tags["series"], tags)
            logger.debug("DATASET.select(tags) found:\n%s ", matching_series)
            expressions.append(nw.col(matching_series))

        df = nw.from_native(self.data).select(expressions).to_native()
        interval = kwargs.get("interval")
        if interval:
            logger.warning(
                f"DATASET.select: Attempted call with argument 'interval' = {interval} is not supported. --> TO DO!"
            )

        match output:
            case "dataframe" | "df":
                out = df
            case "dataset" | "ds":
                if not new_name:
                    new_name = f"COPY of({self.name} SELECTED by names {names}, pattern: {pattern}, regex: {regex} tags: {tags})"
                out = self.copy(new_name, data=df, **kwargs)
                matching_series_tags = {
                    k: v
                    for k, v in out.tags["series"].items()
                    if k in out.numeric_columns
                }
                out.tags["series"] = matching_series_tags
            case _:
                out = df
        return out

    @no_type_check
    def __getitem__(
        self,
        names_or_tags: str | meta.TagDict | Iterable[str | meta.TagDict] = "",
    ) -> Self | None:
        """Access Dataset.data.columns via Dataset.select on list[column_names | tags ] | pattern | regex.

        Arguments:
            names_or_tags:  Either a string pattern, name, list of names or a dict of tags.

        Returns:
            Self | None

        Raises:
            TypeError: If select() returns another type than Dataset.
        """
        if names_or_tags and isinstance(names_or_tags, str) and "*" in names_or_tags:
            result = self.select(regex=names_or_tags)
            # TODO: s is uncharted territory. Explore behaviour / add test cases!
        elif names_or_tags and isinstance(names_or_tags, str):
            result = self.select(names_or_tags)
        elif names_or_tags and isinstance(names_or_tags, dict):
            t = [names_or_tags]
            result = self.select(tags=t)
        elif names_or_tags and isinstance(names_or_tags, Iterable):
            n = [s for s in names_or_tags if isinstance(s, str)]
            t = [d for d in names_or_tags if isinstance(d, dict)]
            result = self.select(n, tags=t)
            # TODO: ensure test cases cover all permutations (a,b) that can be drawn from [{a},{},'a',''] and [{b},{},'b',''] (expect errors raised for combinations with only empty items).
        elif names_or_tags:
            raise TypeError(
                f"DATASET.__getitem__ received a positional argument of type {type(names_or_tags)}. Required: > str | list[str|dict] with a single name, a list of names, a list of tag dicts. Or even a list with combinations of names and tag dicts."
            )
        else:
            return None

        if isinstance(result, Dataset):
            return result
        else:
            raise TypeError(
                "DATASET.__getitem__ call to .select() did not return a Dataset type."
            )

    def __setitem__(
        self, columns: str | Iterable[str], data: IntoFrame | np.ndarray
    ) -> None:
        """Access Dataset.data.columns via Dataset[ list[column_names]."""
        if columns and isinstance(columns, str):
            columns = [columns]
        elif columns and isinstance(columns, Iterable):
            columns = list(columns)
        else:
            raise TypeError(
                "DATASET.__setitem__ columns should be specified as string or iterable, not {type(columns).__name__}."
            )
        nw_self = nw.from_native(self.nw.to_arrow())

        # if False:
        if isinstance(data, np.ndarray):
            # The array shape must be aligned with the numeric subset self[columns].
            if data.ndim == 1:
                data = data.reshape(-1, 1)

            if len(columns) != data.shape[1]:
                raise ValueError(
                    f"Key has {len(columns)} cols, but NumPy array has {data.shape[1]} cols."
                )
            if len(self) != data.shape[0]:
                raise ValueError(
                    f"Dataset has {len(self)} rows, but NumPy array has {data.shape[0]} rows."
                )

            temp_col_names = [f"__temp_{i}" for i in range(data.shape[1])]
            data_df = nw.from_dict(
                {col_name: data[:, i] for i, col_name in enumerate(temp_col_names)},
                backend=nw_self.implementation,
            )
            expressions = [
                data_df.get_column(temp_name).alias(final_name)
                for temp_name, final_name in zip(temp_col_names, columns, strict=False)
            ]
        else:
            data_to_write = nw.from_native(data, backend=nw_self.implementation).select(
                ncs.numeric()
            )

            # TODO: do more to validate/normalize dtypes here?
            # ... and validate/name/order columns?
            columns_dtypes = {
                k: _nw_normalize_dtype(v)
                for k, v in nw_self.schema.items()
                if k in columns
            }
            expressions = [
                data_to_write.get_column(original_name).alias(new_name).cast(dtype)
                for new_name, original_name, dtype in zip(
                    columns, data_to_write.columns, columns_dtypes, strict=False
                )
            ]

        self.data = nw_self.with_columns(expressions).to_native()

    def __len__(self) -> int:
        """Returns the length of the dataset along the time axis, ie. the number of rows."""
        (length, _) = self.data.shape
        return int(length)

    def plot(self, *args: Any, **kwargs: Any) -> Any:
        """Plot dataset data.

        Convenience wrapper around Dataframe.plot() with sensible defaults.
        """
        import pandas

        df = nw.from_native(self.data).to_pandas()

        if self.data_type.temporality == Temporality.FROM_TO:
            interval_handling = kwargs.pop("interval_handling", "interval").lower()
            match interval_handling:
                case "interval":
                    from_data = df
                    to_data = df
                    from_data["valid_to"] = from_data["valid_from"]
                    df = pandas.concat(
                        [from_data, to_data],
                        axis=0,
                        ignore_index=True,
                    ).sort_values(by=["valid_from", "valid_to"])
                    df.drop(columns=["valid_to"], inplace=True)
                    xlabels = "valid_from"
                case "midpoint":
                    xlabels = "midpoint"
                    df["midpoint"] = df[self.datetime_columns].median(axis=1)
                    df.drop(columns=["valid_from", "valid_to"], inplace=True)

                case _:
                    raise ValueError(
                        "Invalid option for interval_handling. Must be 'from', 'to', 'interval' or 'midpoint'."
                    )
        else:
            xlabels = self.datetime_columns[0]

        logger.debug("DATASET.plot(): x labels = %s", xlabels)
        logger.debug(f"Dataset.plot({args!r}, {kwargs!r}) x-labels {xlabels}")

        return df.plot(
            xlabels,
            *args,
            legend=len(self.nw.columns) < 9,
            title=self.name,
            figsize=(12, 4),
            **kwargs,
        )

    def vectors(self, pattern: str = "") -> None:
        """Get vectors with names equal to column names from Dataset.data.

        Args:
            pattern (str): Optional pattern for simple selecting of column names containing pattern. Defaults to ''.

        .. warning:: Caution!
            This (re)assigns variables in the scope of the calling function by way of stack inspection and hence risks of reassigning objects, functions, or variables if they happen to have the same name.

        """
        import inspect

        stack = inspect.stack()
        locals_ = stack[1][0].f_locals

        for col in self.nw.columns:
            if col.__contains__(pattern):
                cmd = f"{col} = self.nw['{col}']"
                logger.debug(cmd)
                # the original idea was running (in caller scope)
                # exec(cmd)
                locals_[col] = self.nw[col]

    def groupby(
        self,
        freq: str,
        func: str = "auto",
        *args: Any,
        **kwargs: Any,
    ) -> Self:
        """Group dataset data by specified frequency and function.

        Returns a new Dataset.
        """
        datetime_columns = list(
            set(self.nw.columns) & {"valid_at", "valid_to", "valid_from"}
        )

        # works for datetime_columns = "valid_at", untested for others
        # TODO: add support for ["valid_from", "valid_to"]
        period_idx = period_index(self.data[datetime_columns[0]], freq=freq)
        logger.debug("DATASET %s: period index\n%s.", self.name, period_idx)

        # Fix for case when **kwargs contains numeric_only
        if "numeric_only" in kwargs:
            kwargs.pop("numeric_only")
        numeric_only_value = True

        match func:
            case "mean":
                out = self.data.groupby(period_idx).mean(
                    *args, numeric_only=numeric_only_value, **kwargs
                )
            case "sum":
                out = self.data.groupby(period_idx).sum(
                    *args, numeric_only=numeric_only_value, **kwargs
                )
            case "auto":
                # TODO: QA on exact logic / use "real" metadata
                # in particular, how to check meta data and blend d1 and df2 values as appropriate
                # (this implementation is just to show how it can be done)
                # QUESTION: do we need a default for "other" series / what should it be?
                df1 = self.data.groupby(period_idx).mean(
                    *args, numeric_only=numeric_only_value, **kwargs
                )
                logger.debug(f"groupby\n{df1}.")

                df2 = (
                    self.data.groupby(period_idx)
                    .sum(*args, numeric_only=numeric_only_value, **kwargs)
                    .select(regex="mendgde|volum|vekt")
                )
                logger.debug(f"groupby\n{df2}.")

                df1[df2.columns] = df2[df2.columns]

                out = df1
                logger.debug(f"groupby\n{out}.")
                logger.debug(f"DATASET {self.name}: groupby\n{out}.")

        new_name = f"({self.name}.groupby({freq},{func})"

        return self.__class__(
            name=new_name,
            data_type=self.data_type,
            as_of_tz=self.as_of_utc,
            data=out,
        )

    def resample(
        self,
        freq: str,
        func: F | str,
        *args: Any,
        **kwargs: Any,
    ) -> Self:
        """Alter frequency of dataset data."""
        # TODO: have a closer look at dates returned for last period when upsampling
        # df = self.data.set_index(self.datetime_columns)
        df = self.data.set_index(self.datetime_columns).copy()
        match func:
            case "min":
                out = df.resample(freq).min()
            case "max":
                out = df.resample(freq).max()
            case "sum":
                out = df.resample(freq).sum()
            case "mean":
                out = df.resample(freq).mean()
            case "ffill":
                out = df.resample(freq).ffill()
            case "bfill":
                out = df.resample(freq).bfill()
            case _:
                out = df.resample(freq, *args, **kwargs).apply(func)

        new_name = f"new set:[{self.name}.resampled({freq}, {func}]"
        return self.__class__(
            name=new_name,
            data_type=self.data_type,
            as_of_tz=self.as_of_utc,
            data=out,
        )

    # TODO: rethink identity: is / is not behaviour
    # def identical(self, other:Self) -> bool:
    #     # check_data = self.__eq__(other:Self)
    #     # return all(check_defs) and check_data.all()
    #     return self.__dict__ == other.__dict__

    def all(self) -> bool:
        """Check if all values in series columns evaluate to true."""
        return bool(np.all(self.numeric_array()))
        # wrong: all(self.data) -- why?

    def any(self) -> bool:
        """Check if any values in series columns evaluate to true."""
        return bool(np.any(self.numeric_array()))

    @property
    def boolean_columns(self) -> list[str]:
        """Get names of all numeric series columns (ie columns that are not datetime)."""
        return list(nw.from_native(self.data).select(ncs.boolean()).columns)
        # replaces: return [c for c in self.data.columns if c not in self.datetime_columns]

    @property
    def datetime_columns(self) -> list[str]:
        """Get names of applicable datetime columns (as_of, valid_at, valid_from, valid_to)."""
        return list(nw.from_native(self.data).select(ncs.datetime()).columns)
        # no need to check against column names (for current data type)?
        # intersect = set(nw.from_native(self.data).columns) & {"valid_at", "valid_from", "valid_to"}
        # return sorted(list(intersect))

    @property
    def numeric_columns(self) -> list[str]:
        """Get names of all numeric series columns (ie columns that are not datetime)."""
        return sorted(nw.from_native(self.data).select(ncs.numeric()).columns)

    def numeric_array(self, series: str | list[str] = "") -> NDArray:
        """Get the data of numeric series columns in matrix format as a Numpy NDArray.

        This will omit datetime columns, hence is convenient for linear algebra operations.
        Optionally, series names can be provided to get a subset of columns.
        """
        expr = [ncs.numeric() | ncs.boolean()]
        if series:
            expr.append(nw.col(series))
        np_array = nw.from_native(self.data).select(expr).to_numpy()
        return cast("NDArray", np_array)

    @property
    def np(self) -> NDArray:
        """Returns Dataset.data as a (new) Numpy NDArray.

        This is a shorthand for `Dataset.nw.to_numpy()`.
        It is different from the `numeric_array` in that it includes date columns.
        similar shorhands are `pd` for `.to_pandas()`, `pl` for `to_polars(), `pa`for `.to_arrow()`.
        See also 'Dataset.numeric_array()' that returns a numeric matrix omitting the date columns.
        """
        return self.nw.to_numpy()
        # ... returns output ddtype = object, but this is a Narwhals oversight
        # return self.__array__()

    @property
    def nw(self) -> Frame:
        """Returns Dataset.data as a (new) Narwhals frame.

        This allows making use not only of the functionality of Narwhals https://narwhals-dev.github.io/narwhals/api-reference/,
        but also that of other supported libraries through .to_pandas(), to_polars(),to_numpy(),.to_arrow().
        For numpy, see also 'Dataset.numeric_array()' that returns a numeric matrix omitting the date columns.
        """
        return cast(Frame, nw.from_native(self.data))

    @property
    def pa(self) -> pa.Table:
        """Returns Dataset.data as a (new) Pyarrow Table.

        This is a shorthand for `Dataset.nw.to_arrow()`.
        Similar shorhands are `np` for `.to_numpy()`, `pl` for `to_polars(), `pa`for `.to_arrow()`.
        """
        return self.nw.to_arrow()

    @property
    def pd(self) -> pd.DataFrame:
        """Returns Dataset.data as a (new) Pandas dataframe.

        This is a shorthand for `Dataset.nw.to_pandas()`.
        Similar shorhands are `np` for `.to_numpy()`, `pl` for `to_polars(), `pa`for `.to_arrow()`.
        """
        return self.nw.to_pandas()

    @property
    def pl(self) -> pl.DataFrame:
        """Returns Dataset.data as a (new) Polars dataframe.

        This is a shorthand for `Dataset.nw.to_polars()`.
        Similar shorhands are `np` for `.to_numpy()`, `pd` for `to_pandas(), `pa`for `.to_arrow()`.
        """
        return self.nw.to_polars()

    @no_type_check
    def math(
        self,
        other: Self | IntoFrame | IntoSeries | int | float | None,
        func,  # noqa: ANN001
    ) -> Any:
        """Generic helper making math functions work on numeric, non date columns of dataframe to dataframe, matrix to matrix, matrix to vector and matrix to scalar.

        Although the purpose was to limit "boilerplate" for core linear algebra functions, it also extend to other operations that follow the same differentiation pattern.

        Args:
            other (dataframe | series | matrix | vector | scalar ): One (or more?) pandas (polars to come) dataframe or series, numpy matrix or vector or a scalar value.
            func (_type_): The function to be applied as `self.func(**other:Self)` or (in some cases) with infix notation `self f other`. Note that one or more date columns of the self / lefthand side argument are preserved, ie data shifting operations are not supported.

        Raises:
            ValueError: "Unsupported operand type"
            ValueError: "Incompatible shapes."

        Returns:
            Any:   Depending on the inputs: A new dataset / vector / scalar with the result.
            For datasets, the name of the new set is derived from inputs and the functions applied.
            If 'other' is not recognized, the 'NotImplemented' Singleton is returned so that Python can invoke  '__r<method>__' of other class.
        """
        num_cols = self.numeric_columns
        out = self.copy()

        if isinstance(other, Dataset):
            logger.debug(
                f"DATASET {self.name}: .math({self.name}.{func.__name__}(Dataset({other.name}))."
            )
            other_name = other.name
            # ValueError: operands could not be broadcast together with shapes (0,0) (4,3)
            # observed here for 'test_algebra_expression_with_multiple_dataset' -->
            # print(f"DATASET.math({func.__name__})\n=======\n{str(self)=} \n---\n{str(other)=}\n=======")
            result = func(self.numeric_array(), other.numeric_array())
            out[num_cols] = result
            if isinstance(self.as_of_utc, datetime) and isinstance(
                other.as_of_utc, datetime
            ):
                out.as_of_utc = max(self.as_of_utc, other.as_of_utc)
            elif isinstance(self.as_of_utc, datetime):
                out.as_of_utc = self.as_of_utc
            elif isinstance(other.as_of_utc, datetime):
                out.as_of_utc = other.as_of_utc
            else:
                out.as_of_utc = None

        elif other is None:
            logger.debug(f"DATASET {self.name}: .math({self.name}.{func.__name__}).")
            result = func(self.numeric_array())
            out[num_cols] = result
            other_name = ""

        elif is_df_like(other):
            other = nw.from_native(other).select(num_cols).to_numpy()
            out[num_cols] = func(self.numeric_array(), other)
            other_name = "df"

        # TODO: Add coverage elif isinstance(other, pd.Series):
        elif isinstance(other, nw.Series):
            raise NotImplementedError(
                "Adding support for Dataset data (matrix) operations with datasframe Series (vectors) is on the todo-list! Notify devs to increase priority."
            )

        elif isinstance(other, int | float):
            result = func(self.numeric_array(), other)
            out[num_cols] = result
            other_name = str(other)

        elif isinstance(other, np.ndarray):
            # TODO: this needs more thorough testing!
            # Compare shape of the ndarray against the numeric_columns of self.data. There are up to 3 accepted cases (depending on the operation):
            #  * matrix;         shape = (data.numeric.rows, data.numeric.columns)
            #  * column vector;  shape = (data.numeric.rows, 1)
            #  * row vector;     shape = (1, data.numeric.columns)
            #
            if other.ndim == 1 and (
                other.shape[0] == len(self.data)
                or other.shape[0] == len(self.data.columns)
            ):
                out[num_cols] = func(self.numeric_array(), other)
            else:
                raise ValueError(
                    f"Incompatible shapes for element-wise {func.__name__}"
                )
            other_name = "ndarray"
        else:
            # So that __r<op>__ of other object may be called, instead of:
            # raise ValueError(f"Unsupported operand type: {type(other)}.") -->
            return NotImplemented

        new_name = f"({self.name}.{func.__name__}.{other_name})"
        out.rename(new_name)
        out.lineage = new_name

        logger.debug(f"DATASET.math: {new_name}\n\t{out.data.shape}")
        # print( f"DATASET.math: {new_name}\n\t{out.data.shape=}")
        # TODO: Should also update series names and set/series tags
        return out

    def __add__(self, other: Self | IntoFrame | IntoSeries | int | float) -> Any:
        """Add two datasets or a dataset and a dataframe, numpy array or scalar.

        The operation is performed elementwise using Numpy broadcast rules if dimensions does not match. See np.add.
        """
        return self.math(other, np.add)

    def __radd__(self, other: Self | IntoFrame | IntoSeries | int | float) -> Any:
        """Right add handles cases where a Dataset is added to a non-dataset object.

        The operation is performed elementwise using Numpy broadcast rules if dimensions does not match. See np.add.
        """
        return self.__add__(other)

    def __sub__(self, other: Self | IntoFrame | IntoSeries | int | float) -> Any:
        """Subtract two datasets or a dataset and a dataframe, numpy array or scalar.

        The operation is performed elementwise using Numpy broadcast rules if dimensions does not match. See np.sub.
        """
        return self.math(other, np.subtract)

    def __rsub__(self, other: Self | IntoFrame | IntoSeries | int | float) -> Any:
        """Right subtract handles cases where a Dataset is subtrd from a non-Dataset.

        The operation is performed elementwise using Numpy broadcast rules if dimensions does not match. See np.sub.
        """
        negative_self = self.__mul__(-1)
        return negative_self.__add__(other)

    def __mul__(self, other: Self | IntoFrame | IntoSeries | int | float) -> Any:
        """Multiply two datasets or a dataset and a dataframe, numpy array or scalar.

        The operation is performed elementwise using Numpy broadcast rules if dimensions does not match. See np.muliply.
        """
        return self.math(other, np.multiply)

    def __rmul__(self, other: Self | IntoFrame | IntoSeries | int | float) -> Any:
        """Right multiply two datasets or a dataset and a dataframe, numpy array or scalar.

        The operation is performed elementwise using Numpy broadcast rules if dimensions does not match. See np.muliply.
        """
        return self.__mul__(other)

    def __matmul__(self, other: Self | IntoFrame | IntoSeries) -> Any:
        """Matrix multiply two datasets or a dataset and a dataframe, numpy array.

        The matmul function implements the semantics of the @ operator defined in PEP 465.
        The operation inherits behavioural characteristicsa from Numpy. See np.matmul for details.

        Raises:
            ValueError
                If the last dimension of x1 is not the same size as the second-to-last dimension of x2.
                If a scalar value is passed in.
        """
        return self.math(other, np.matmul)

    def __rmatmul__(self, other: Self | IntoFrame | IntoSeries | int | float) -> Any:
        """Right matrix multiply two datasets or a dataset and a dataframe, numpy array or scalar.

        The operation is performed elementwise using Numpy broadcast rules if dimensions does not match. See np.matmul.
        """
        return self.__mul__(other)

    def __truediv__(self, other: Self | IntoFrame | IntoSeries | int | float) -> Any:
        """Divide two datasets, or a dataset and a dataframe, numpy array or scalar.

        The operation is performed elementwise using Numpy broadcast rules if dimensions does not match. See np.divide.
        """
        return self.math(other, np.divide)

    def __rtruediv__(self, other: Self | IntoFrame | IntoSeries | int | float) -> Any:
        """Right true divide handles cases when non-dataset objects are divided by datasets.

        Division is performed elementwise using Numpy broadcast rules. See np.divide.

        The operation is performed elementwise using Numpy broadcast rules if dimensions does not match. See np.divide.
        """
        inverse = self.__truediv__(other)
        # TODO: verify that the above always returns floats.
        # If not, we will get weird results when inverting with np.repiprocal.
        # (Also, check behaviour with decimal datatype.)
        return inverse.math(None, np.reciprocal)

    def __floordiv__(self, other: Self | IntoFrame | IntoSeries | int | float) -> Any:
        """Floor divide two datasets or a dataset and a dataframe, numpy array or scalar.

        The operation is performed elementwise using Numpy broadcast rules if dimensions does not match. See np.floordivide.
        """
        return self.math(other, np.floor_divide)

    def __rfloordiv__(self, other: IntoFrame | IntoSeries | int | float) -> Any:
        """Right floor divide handles cases when a dataframe, numpy array or scalar is divded by a dataset.

        Done element wise w. broadcast.

        The operation is performed elementwise using Numpy broadcast rules if dimensions does not match. See np.floordivide.
        """
        inverse = self.__floordiv__(other)
        return inverse.math(None, np.reciprocal)

    def __pow__(self, other: Self | IntoFrame | IntoSeries | int | float) -> Any:
        """Power of two datasets or a dataset and a dataframe, numpy array or scalar.

        The operation is performed elementwise using Numpy broadcast rules if dimensions does not match. See np.power.
        """
        return self.math(other, np.power)

    def __rpow__(self, other: IntoFrame | IntoSeries | int | float) -> Any:
        """Right power of two datasets or a dataset and a dataframe, numpy array or scalar.

        The operation is performed elementwise using Numpy broadcast rules if dimensions does not match. See np.power.
        """
        # raise NotImplementedError("Coming soon!")
        # self = x, other = a
        # y = a ** x (elementwise for x)
        # y = (e ** x) ** ln(a) ==>
        exp_self = self.math(
            None, np.exp
        )  # .math with None needs more testing / consider Numpy ufunc implementation
        ln_a = np.log(other)  # may handling of input type variations?
        return exp_self.__pow__(ln_a)

    def __mod__(self, other: Self | IntoFrame | IntoSeries | int | float) -> Any:
        """Modulo of two datasets or a dataset and a dataframe, numpy array or scalar.

        The operation is performed elementwise using Numpy broadcast rules if dimensions does not match. See np.mod.
        """
        return self.math(other, np.mod)

    def __rmod__(self, other: Self | IntoFrame | IntoSeries | int | float) -> Any:
        """Right modulo of two datasets or a dataset and a dataframe, numpy array or scalar.

        The operation is performed elementwise using Numpy broadcast rules if dimensions does not match. See np.mod.
        """
        return NotImplemented  # self.__mod__(other)

    def __neg__(self) -> Any:
        """Negative values for dataset.

        The operation returns a new dataset object. See np.negative.
        """
        return self.math(None, np.negative)

    def __pos__(self) -> Any:
        """Positive values for dataset.

        The operation returns a new dataset object. See np.positive.
        """
        return self.math(None, np.positive)

    def __abs__(self) -> Any:
        """Absolute values for dataset.

        The operation returns a new dataset object. See np.absolute.
        """
        return self.math(None, np.absolute)

    def __invert__(self) -> Any:
        """Invert the sign of values in the dataset.

        The operation is performed elementwise. See np.negative.
        """
        return self.math(-1, np.multiply)

    @no_type_check
    def __eq__(
        self, other: object
    ) -> Any:  # Self | IntoFrame | IntoSeries | int | float) -> Any:
        """Check equality of two datasets or one dataset and a dataframe, numpy array or scalar.

        Note that this is a pure value comparison, and thus different from comparing dataset identity as reflected in the metadata of two dataset instances.

        The operation is performed elementwise using Numpy broadcast rules if dimensions does not match. See np.equal.
        """
        return self.math(other, np.equal)

    def __ne__(
        self, other: Any
    ) -> Any:  # Self | IntoFrame | IntoSeries | int | float) -> Any:
        # def __ne__(self, other:Self | IntoFrame | IntoSeries | int | float) -> Any:
        """Check inequality of two datasets or one dataset and a dataframe, numpy array or scalar.

        Note that this is a pure value comparison, and thus different from comparing dataset identity as reflected in the metadata of two dataset instances.

        The operation is performed elementwise using Numpy broadcast rules if dimensions does not match. See np.not_equal.
        """
        return self.math(other, np.not_equal)

    def __gt__(self, other: Self | IntoFrame | IntoSeries | int | float) -> Any:
        """Check greater than for two datasets or a dataset and a dataframe, numpy array or scalar.

        The operation is performed elementwise using Numpy broadcast rules if dimensions does not match. See np.greater.
        """
        return self.math(other, np.greater)

    def __lt__(self, other: Self | IntoFrame | IntoSeries | int | float) -> Any:
        """Check less than for two datasets or a dataset and a dataframe, numpy array or scalar.

        The operation is performed elementwise using Numpy broadcast rules if dimensions does not match. See np.less.
        """
        return self.math(other, np.less)

    def __ge__(self, other: Self | IntoFrame | IntoSeries | int | float) -> Any:
        """Check greater than or equal for two datasets or a dataset and a dataframe, numpy array or scalar.

        The operation is performed elementwise using Numpy broadcast rules if dimensions does not match. See np.greater_equal.
        """
        return self.math(other, np.greater_equal)

    def __le__(self, other: Self | IntoFrame | IntoSeries | int | float) -> Any:
        """Check less than or equal for two datasets or a dataset and a dataframe, numpy array or scalar.

        The operation is performed elementwise using Numpy broadcast rules if dimensions does not match. See np.less_equal.
        """
        return self.math(other, np.less_equal)

    def isclose(self, other: Self | IntoFrame | IntoSeries | int | float) -> Any:
        """Check if two datasets or a dataset and a dataframe, numpy array or scalar are nearlly equal.

        The operation is performed elementwise using Numpy broadcast rules if dimensions does not match. See np.isclose.
        """
        return self.math(other, np.isclose)

    def __repr__(self) -> str:
        """Returns a machine readable string representation of Dataset, ideally sufficient to recreate object."""
        if isinstance(self.as_of_utc, datetime):
            as_of = f'"{self.as_of_utc.isoformat()}"'
        else:
            as_of = None
        return f'Dataset(name="{self.name}", repository="{self.repository}", data_type={self.data_type!r}, as_of_tz={as_of})'

    def __str__(self) -> str:
        """Returns a human readable string representation of the Dataset."""
        return str(
            {
                "name": self.name,
                "data_type": str(self.data_type),
                "as_of_utc": self.as_of_utc,
                "repository": self.repository,
                "series": str(self.series),
                "data": self.data.shape,
            }
        )

    def __array__(
        self,
        dtype: type | DTypeLike | None = None,
        copy: bool = True,
    ) -> NDArray:
        """Returns the series in the Dataset as a Numpy array, adhering to the array protocol.

        The function returns `numeric_array` unless `dtype` is specified.
        Then it follows the Narwhals implementation.
        See also Dataset.np for including datetime columns.
        """
        if dtype is None:
            # Numeric only is better fit for the most likely purpose: Numpy for mathematic.
            return self.numeric_array()
        else:
            return self.nw.__array__(dtype=dtype, copy=copy)

    def __arrow_c_stream__(
        self,
        requested_schema: Any = None,  # WTF? mypy fails for pa.Schema
    ) -> Any:  # WTF? mypy fails for pa.RecordBatchReader
        """Implements the Arrow C Data Interface.

        This allows PyArrow and other Arrow compatible libraries or even other programming languages to collect data from a Dataset object.
        """
        return self.nw.__arrow_c_stream__(requested_schema=requested_schema)

    def __dataframe__(
        self,
        nan_as_null: bool = False,
        allow_copy: bool = True,
    ) -> Any:
        """Implements the Dataframe Interchange Protocol."""
        return self.pa.__dataframe__(
            nan_as_null=nan_as_null,
            allow_copy=allow_copy,
        )

    def aggregate(
        self,
        attributes: list[str],
        taxonomies: list[int | meta.Taxonomy | dict[str, str] | PathStr],
        functions: list[str | F] | set[str | F],
        sep: str = "_",
    ) -> Self:
        """Aggregate dataset by taxonomy hierarchies.

        Args:
            attributes (list[str]): The attributes to aggregate by.
            taxonomies (list[int | meta.Taxonomy | dict[str, str] | PathStr]):
                Value definitions for `attributes` can be either ::py:class:`meta.Taxonomy` objects,
                or klass_ids, data dictionaries or paths that can be used to retrieve or construct them.
            functions (list[str|F] | set[str|F]): Optional function name (or list) of the function names to apply (mean | count | sum | ...). Defaults to `sum`.
            sep (str): Optional separator used when joining multiple attributes into names of aggregated series. Defaults to '_'.

        Returns:
            Self: A dataset object with the aggregated data.
            If the taxonomy object has hierarchical structure, aggregate series are calculated for parent nodes at all levels.
            If the taxonomy is a flat list, only a single `total` aggregate series is calculated.

        Raises:
            TypeError: If any of the taxonomy identifiere are of unexpected types.

        Examples:
            To calculate 10 and 90 percentiles and median for the dataset `x` where codes from KLASS 157 (energy_balance) distinguishes between series in the set.

            >>> from ssb_timeseries.dataset import Dataset
            >>> from ssb_timeseries.types import SeriesType
            >>> from ssb_timeseries.sample_data import create_df
            >>> from ssb_timeseries.meta import Taxonomy
            >>>
            >>> klass157 = Taxonomy(klass_id=157)
            >>> klass157_leaves = [n.name for n in klass157.structure.root.leaves]
            >>> tag_permutation_space = {"A": klass157_leaves, "B": ["q"], "C": ["z"]}
            >>> series_names: list[list[str]] = [value for value in tag_permutation_space.values()]
            >>> sample_df = create_df(*series_names, start_date="2024-01-01", end_date="2024-12-31", freq="MS",)
            >>> sample_set = Dataset(name="sample_set",
            >>>     data_type=SeriesType.simple(),
            >>>     data=sample_df,
            >>>     attributes=["A", "B", "C"],
            >>> )
            >>>
            >>> def perc10(x):
            >>>     return x.quantile(.1, axis=1, numeric_only=True, interpolation="linear")
            >>>
            >>> def perc90(x):
            >>>     return x.quantile(.9, axis=1, numeric_only=True, interpolation="linear")
            >>>
            >>> percentiles = sample_set.aggregate(["energy_balance"], [157], [perc10, 'median', perc90])
        """
        taxonomy_dict = {}
        for name, t in zip(attributes, taxonomies, strict=False):
            if isinstance(t, meta.Taxonomy):
                obj = t
            elif isinstance(t, int):
                obj = meta.Taxonomy(klass_id=t)
            elif isinstance(t, dict):
                obj = meta.Taxonomy(data=t)
            elif isinstance(t, (str | Path)):
                obj = meta.Taxonomy(path=t)
            else:
                raise TypeError(
                    f"Taxonomy object or valid identifier expected, got {type(t)}"
                )
            taxonomy_dict[name] = obj

        df = nw.from_native(self.data).drop(self.numeric_columns).to_pandas()
        new_series_tags = {}
        permutations = meta.permutations(taxonomy_dict, "parents")
        for p in permutations:
            criteria = {}
            for attr, value in p.items():
                criteria[attr] = taxonomy_dict[attr].leaf_nodes(value)
            output_series_name = sep.join(p.values())  # move into func loop?
            leaf_node_subset = self.select(tags=criteria, output="df")
            for func in functions:
                if isinstance(func, str):
                    new_col_name = f"{func}({output_series_name})"
                    func_name = func
                elif isinstance(func, list):
                    new_col_name = f"{func[0]}{func[1]}({output_series_name})"
                    func_name = f"{func[0]}{func[1]}"
                else:
                    new_col_name = f"{func.__name__}({output_series_name})"
                    func_name = func.__name__
                df[new_col_name] = column_aggregate(leaf_node_subset, func)
                lineage_info = {
                    "criteria": criteria,
                    "input": leaf_node_subset.columns,
                    "output": new_col_name,
                    "function": func_name,
                }
                new_series_tags[new_col_name] = lineage_info

        out = self.copy(f"{self.name}.{functions}", data=df)
        # TODO: the content of 'calculations' must be properly placed,
        # but for now it goes here.
        out.tags["calculations"] = new_series_tags
        return out

    def moving_average(
        self,
        start: int = 0,
        stop: int = 0,
        nan_rows: str = "return",
    ) -> Self:
        """Returns a new Dataset with moving averages for all series.

        The average is calculated over a time window defined by `from` and `to` period offsets.
        Negative values denotes periods before current, positive after.
        Both default to 0, ie the current period; so at least one of them should be used.

        >>> x.moving_average(start= -3, stop= -1) # xdoctest: +SKIP
        signifies the average over the three periods before (not including the current).

        Offset parameters will overflow the date range at the beginning and/or end,
        Moving averages can not be calculated.

        Set the parameter `nans` to control the behaviour in such cases:
        'return' to return rows with all NaN values (default).
        'remove' to remove these rows from both ends.

        TO DO: Add parameter to choose returned time window?
        TO DO: Add more NaN handling options?
        TO DO: Add parameter to ensure/alter sampling frequency before calculating.
        """
        n = stop - start + 1
        numbers = self.numeric_array()
        rows, columns = numbers.shape

        r = np.array(range(rows))
        r_from = r + start
        r_to = r + stop
        r_intersect = np.intersect1d(r_from, r_to)

        nans = np.ones((n, columns)) * np.nan
        numbers_ext = np.append(numbers, nans, 0)

        cumsums = np.cumsum(numbers_ext, axis=0)
        zeros = np.zeros((1, columns))
        diffs = cumsums[r_to, :] - np.append(zeros, cumsums, 0)[r_from, :]

        averages = diffs / n
        out = self.copy(f"{self.name}.mov_avg({start},{stop})")
        out[self.numeric_columns] = averages[0:rows, :]

        # how to handle nans? choice between multiple strategies;
        # ... + choice whether to distinguish between beginning/end/middle?
        # - just return them (keep all rows) --> default: 'return'
        # - remove --> 'remove' 'remove_beginning' 'remove_end'
        # - calculate something
        match nan_rows:
            case "remove":
                result_rows = r_intersect
            case "remove_beginning":
                result_rows = np.array(range(min(r_intersect), rows))
            case "remove_end":
                result_rows = np.array(range(0, max(r_intersect)))
            case "return":
                result_rows = r
            case _:
                raise (
                    ValueError(
                        f"Received {nan_rows=}; allowed values include return | remove | ... (See the docs for more.) "
                    )
                )
            # - return some value, use method:
            #   - ignore nans when calculating avg
            #   - impute, repeat/prepend first/interpolate
            #   - estimation? henderson or other
            #   - retrieve more values?

        out.data = out.data.take(result_rows)
        # TODO: update the metadata
        return out


def column_aggregate(df: FrameT, method: str | F) -> Any:
    """Helper function to calculate aggregate over dataframe columns."""
    logger.debug("DATASET.column_aggregate '%s' over columns:\n%s", method, df.columns)
    nw_df = nw.from_native(df)

    # the following is not pretty, but is left as is to simplify the transition away from pandas
    # a better approach: return nw.Expr for methods  --> TODO!
    pd_df = nw_df.to_pandas()

    if isinstance(method, F):  # type: ignore
        out = method(pd_df)  # type: ignore[operator]
    else:
        match method:
            case "mean" | "average":
                out = pd_df.mean(axis=1, numeric_only=True)
            case "min" | "minimum":
                out = pd_df.min(axis=1, numeric_only=True)
            case "max" | "maximum":
                out = pd_df.max(axis=1, numeric_only=True)
            case "count":
                out = pd_df.count(axis=1, numeric_only=True)
            case "sum":
                out = pd_df.sum(axis=1, numeric_only=True)
            case "median":
                out = pd_df.quantile(0.5, axis=1, numeric_only=True)
            case ["quantile", *params] | ["percentile", *params]:
                if len(params) > 1:
                    interpolation: str = params[1]
                else:
                    interpolation = "linear"
                quantile = float(params[0])
                if quantile > 1:
                    quantile /= 100
                out = pd_df.quantile(
                    quantile,
                    axis=1,
                    numeric_only=True,
                    interpolation=interpolation,
                )
            case _:
                raise NotImplementedError(
                    f"Aggregation method '{method}' is not implemented (yet)."
                )
    return nw.new_series(
        name=f"{method}({pd_df.columns})",
        values=out,
        backend=nw_df.implementation,
    )


def search(
    equals: str = "",
    contains: str = "",
    pattern: str = "*",
    as_of_tz: datetime = None,
    repository: str = "",
    require_unique: bool = False,
    # ) -> list[old_io.SearchResult] | Dataset | None:
) -> list | Dataset | None:
    """Search for datasets by name matching pattern.

    Returns:
         list[io.SearchResult] | Dataset | list[None]: The dataset for a single match, a list for no or multiple matches.

    Raises:
        ValueError: If `require_unique = True` and a unique result is not found.
    """
    # TODO: REFACTOR (using catalog?)
    # from .io import simple as old_io
    # found = old_io.find_datasets(
    #    pattern=pattern,
    #    repository=repository,
    # )
    # logger.debug(
    #    "DATASET.search for '%s'\nin repositories\n%s\nreturned:\n%s",
    #    pattern,
    #    repository,
    #    found,
    # )
    from .catalog import get_catalog

    c = get_catalog()
    found = c.datasets(
        equals=equals,
        contains=contains,
        pattern=pattern,
        repository=repository,
    )
    number_of_results = len(found)

    if number_of_results == 1:
        tags = found[0].object_tags
        stype = SeriesType(tags["versioning"], tags["temporality"])
        return Dataset(
            name=tags["name"],
            data_type=stype,
            as_of_tz=as_of_tz,
        )
    elif number_of_results == 0:
        return None
    elif require_unique:
        raise ValueError(
            f"Search for '{pattern}' returned:\n{number_of_results} results when exactly one was expected:\n{found}",
        )
    else:
        return found


def _has_auto_tag_information(ds: Dataset, caller_workspace: dict) -> bool:
    """Checks if a dataset has the information required to run .series_names_to_tags()."""
    autotag_config_exists = hasattr(ds, "auto_tag_config")
    attr_var_exists = "attributes" in caller_workspace
    if attr_var_exists and caller_workspace["attributes"]:
        return True
    elif (
        autotag_config_exists
        and hasattr(ds["auto_tag_config"], "attributes")
        and ds["auto_tag_config"]["attributes"]
    ):
        return not is_empty(ds.data)
        # (no need to check separator or regex)
    else:
        return False


def _nw_normalize_dtype(t: IntoDType) -> IntoDType:
    """Normalize datatypes."""
    match t:
        case nw.Int8 | nw.Int16 | nw.Int32:
            out = nw.Int64
        case nw.Float32:
            out = nw.Float64
        case nw.Date:
            out = nw.Datetime
        case _:
            out = t
    return out



# ruff: noqa   #NOSONAR
"""
Helper module for date and time utility functions.

Notable examples include converting between UTC and local time, standardised string formats for file names, and (planned for future use) intervals.
"""

from __future__ import annotations

from datetime import datetime, tzinfo
from typing import Any, Iterable
from typing import Literal
from typing import TypeAlias
from typing import cast
from zoneinfo import ZoneInfo
from multipledispatch import dispatch

from dateutil import parser
from narwhals.typing import IntoFrameT, FrameT, IntoSeriesT
import narwhals as nw
import narwhals.selectors as ncs
from pandas import PeriodIndex
import pyarrow as pa
from .logging import logger


# TODO: align these / control by configuration
MAX_TIME_PRECISION: str = "second"
DEFAULT_TIMESPEC: str = "seconds"
NW_DEFAULT_TIME_UNIT: Literal["ns", "us", "ms"] = "ns"

# PyArrow related constants to ensure consistent typing for timestamp columns
PA_TIMESTAMP_UNIT: Literal["ns", "us", "ms", "s"] = "ns"
PA_TIMESTAMP_TZ: str = "UTC"

DEFAULT_TZ = ZoneInfo("Europe/Oslo")  # Will shift between CET and CEST
CET = ZoneInfo("CET")
UTC = ZoneInfo("UTC")
TimeZone: TypeAlias = ZoneInfo | str | None


def date_utc(some_date: datetime | str | None, **kwargs) -> datetime:
    """Convert datetime or date string to UTC.

    If date has no timezone information, the data is assumed to be in default timezone (CET).

    The output will be rounded to the precision specified by kwarg 'rounding'.
    """
    if some_date is None or some_date == "":
        return date_round(now_utc())

    else:
        dt_type = ensure_datetime(some_date)
        tz_aware = ensure_tz_aware(dt_type)
        correct_tz = tz_aware.astimezone(tz=UTC)
        return date_round(correct_tz, **kwargs)


def date_local(some_date: datetime | str, **kwargs) -> datetime:
    """Convert date to default timezone.

    If not configured otherwise the default is Europe/Oslo which provides automatic shifts between CET and CEST.
    The output can be rounded to the precision specified by kwarg 'rounding'.
    """
    dt_type = ensure_datetime(some_date, tz=DEFAULT_TZ)
    return date_round(dt_type, **kwargs)


def date_cet(some_date: datetime | str, **kwargs) -> datetime:
    """Convert date to time_zone Europe/Oslo which provides automatic shifts between CET and CEST.

    The output can be rounded to the precision specified by kwarg 'rounding'.
    """
    dt_type = ensure_datetime(some_date, tz=CET)
    return date_round(dt_type, **kwargs)


def date_round(d: datetime, **kwargs) -> datetime:
    """Round date to specified by kwarg 'rounding' or default precision MAX_TIME_PRECISION.

    Rounding can take the values 'none', 'day', 'd', 'hour', 'h', 'minute', 'min', 'm', 'second', 'sec', or 's'.

    Default precision 'seconds' is used if none is provided.
    """

    if not d:
        return d

    rounding = kwargs.get("rounding", MAX_TIME_PRECISION)
    match rounding.lower():
        case "day" | "d":
            out = d.replace(hour=0, minute=0, second=0, microsecond=0)
        case "hour" | "h":
            out = d.replace(minute=0, second=0, microsecond=0)
        case "minute" | "min" | "m":
            out = d.replace(second=0, microsecond=0)
        case "second" | "sec" | "s":
            out = d.replace(microsecond=0)
        case "microsecond" | "mic" | "u" | "none":
            out = d
    return out


def ensure_datetime(some_date_representation: Any, **kwargs) -> datetime:
    """Make sure that we are dealing with a datetime object, convert if possible.

    If input is None or empty strings will be converted to now_utc().
    """
    if isinstance(some_date_representation, datetime):
        return some_date_representation
    elif some_date_representation is None or some_date_representation == "":
        date_as_dt = now_utc(**kwargs)
    else:
        try:
            date_as_dt = some_date_representation.to_datetime()
        except (ValueError, TypeError, AttributeError):
            date_as_dt = parser.parse(some_date_representation)
        return date_as_dt
    return date_as_dt


def ensure_tz_aware(some_date: datetime) -> datetime:
    """Make sure that our datetime object is timezone aware.

    Assume CET if timezone information is missing.
    """
    if is_tz_naive(some_date):
        logger.debug(
            "DATE_UTC catched a date without timezone info. This will become an error later. Assuming CET."
        )
        return some_date.replace(tzinfo=DEFAULT_TZ)
    else:
        return some_date


def is_tz_aware(d: datetime) -> bool:
    return d.tzinfo is not None and d.tzinfo.utcoffset(d) is not None


def is_tz_naive(d: datetime) -> bool:
    return d.tzinfo is None or d.tzinfo.utcoffset(d) is None


def now_utc(**kwargs) -> datetime:
    """Return now in UTC."""
    t = datetime.now(tz=UTC)
    return date_round(t, **kwargs)


def now_cet(**kwargs) -> datetime:
    """Return now in CET."""
    t = datetime.now(tz=CET)
    return date_round(t, **kwargs)


def utc_iso(d: Any, timespec: str = DEFAULT_TIMESPEC) -> str:
    """Convert date to UTC and return as an ISO formatted string."""
    return date_utc(d).isoformat(timespec=timespec)


def utc_iso_no_colon(d: datetime, timespec: str = DEFAULT_TIMESPEC) -> str:
    """Convert date to UTC and return as an ISO formatted string without the colons."""
    return utc_iso(d, timespec=timespec).replace(":", "")


def prepend_as_of(
    df: nw.typing.IntoFrameT, as_of: datetime | None
) -> nw.typing.IntoFrameT:
    """Prepend column 'as_of' to dataframe."""
    nw_df = cast(nw.DataFrame, nw.from_native(df))

    if as_of is not None:
        as_of_value = date_utc(as_of)
        return nw_df.with_columns(nw.lit(as_of_value).alias("as_of")).to_native()
    else:
        # Create a column of typed nulls to avoid type inference errors
        nw_df = cast(nw.DataFrame, nw.from_native(df))
        as_of_nulls_series = nw.lit(None).cast(
            nw.Datetime(time_unit=PA_TIMESTAMP_UNIT, time_zone=PA_TIMESTAMP_TZ)
        )
        return nw_df.with_columns(as_of=as_of_nulls_series).to_native()


def local_timezone() -> ZoneInfo:
    """Return the local timezone of the computer."""
    return cast(ZoneInfo, datetime.now().astimezone().tzinfo)


def validate_timezone(tz: TimeZone = "") -> str:
    """Return a valid time zone as string or 'DEFAULT_TZ' for the empty string or 'None'."""
    if not tz:
        tz = DEFAULT_TZ
    return str(tz)


def _nw_expr_datelike_to_datetime() -> list[nw.Expr]:
    """Returns a Narwhals expression to transform all Date columns to Datetime."""
    expressions = [cast(nw.Expr, ncs.by_dtype(nw.Object, nw.Date).cast(nw.Datetime))]
    return expressions


def _nw_expr_datetime_time_unit(
    schema=nw.Schema, time_unit: Literal["ns", "us", "ms", "s"] = NW_DEFAULT_TIME_UNIT
) -> list[nw.Expr]:
    """Returns a Narwhals expression to transform all Date columns to Datetime."""
    expressions = []
    for col_name, dtype in schema.items():
        if dtype in (nw.Date, nw.Datetime):
            expressions.append(
                nw.col(col_name).cast(
                    nw.Datetime(time_unit=time_unit, time_zone=dtype.time_zone)
                )
            )
    return expressions


def _nw_expr_tz_localize(schema: nw.Schema, target_tz: TimeZone) -> list[nw.Expr]:
    """Returns a list of expressions to localize all tz anive Datetime columns to target time zone."""
    expressions = _nw_expr_datelike_to_datetime()
    tz = validate_timezone(target_tz)
    for col_name, dtype in schema.items():
        if (
            dtype in (nw.Date, nw.Datetime)
            and getattr(dtype, "time_zone", None) is None
        ):
            expressions.append(nw.col(col_name).dt.replace_time_zone(tz))
    return expressions


def _nw_expr_tz_convert(schema: dict, target_tz: TimeZone) -> list[nw.Expr]:
    """Scans the schema and returns a list of expressions for Datetime columns
    that are in a different timezone than the target.
    """
    expressions = []
    tz = validate_timezone(target_tz)
    for col_name, dtype in schema.items():
        if (
            dtype in (nw.Date, nw.Datetime)
            and dtype.time_zone is not None
            and dtype.time_zone != tz
        ):
            expressions.append(nw.col(col_name).dt.convert_time_zone(tz))
    return expressions


def datetime_time_unit(
    df: IntoFrameT, time_unit: Literal["ns", "us", "ms", "s"] = NW_DEFAULT_TIME_UNIT
) -> IntoFrameT:
    """Ensure all datetime columns of a dataframe use the same time unit."""
    nw_df = cast(nw.DataFrame, nw.from_native(df))
    expression = _nw_expr_datetime_time_unit(nw_df.schema, time_unit=time_unit)
    return nw_df.with_columns(expression).to_native()


def datelike_to_datetime(
    df: IntoFrameT,
) -> IntoFrameT:
    """Convert all datelike columns of a dataframe to Narwhals Datetime.

    Any timezone information (or lack there of) is passed through without explicit transformation.
    Note that this does not guarantee time zone information is completely untouched.
    Implicit localizations (typically to UTC) may still be triggered by Narwhals type transitions for some backends.

    """
    nw_df = cast(nw.DataFrame, nw.from_native(df))
    expression = _nw_expr_datelike_to_datetime()
    return nw_df.with_columns(expression).to_native()


def datetime_localize(df: IntoFrameT, target_tz: TimeZone = "") -> IntoFrameT:
    """Ensure all datetime columns of a dataframe are timezone aware.

    Columns without timezone information are localized using 'target_tz' parameter if provided, otherwise falling back to default.

    If the dataframe contains datelike columns of types other than Datetime, ie. Date, Object or string representations, see the twin function 'datelike_localize()'
        >>> # xdoctest: +SETUP
        >>> import pandas as pd
        >>> from ssb_timeseries.dates import datelike_to_datetime
        >>> # ------------------------------------------------------------

        >>> df = pd.DataFrame({'time': [
        ...     '2022-01-01 11:30',
        ...     '2022-01-01 12:00',
        ...     '2022-01-01 12:30',
        ...     '2022-01-01 13:00'
        ... ]})
        >>> datetime_localize(datelike_to_datetime(df))
    """
    nw_df = cast(nw.DataFrame, nw.from_native(df))
    expression = _nw_expr_tz_localize(
        nw_df.schema,
        target_tz,
    )
    return nw_df.with_columns(expression).to_native()


def datelike_localize(df: IntoFrameT, target_tz: TimeZone = "") -> IntoFrameT:
    """Convert all datelike columns of a dataframe to timezone aware Narwhals Datetime.

    Equivalent to
        >>> # xdoctest: +SETUP
        >>> import pandas as pd
        >>> from ssb_timeseries.dates import datelike_to_datetime
        >>> # ------------------------------------------------------------

        >>> df = pd.DataFrame({'time': ['2022-01-01','2022-01-01','2022-01-01','2022-01-01']})
        >>> datetime_localize(datelike_to_datetime(df))
    """
    df_with_dt_cols = datelike_to_datetime(df)
    return datetime_localize(df_with_dt_cols, target_tz)


def datetime_convert_timezone(
    df: IntoFrameT,
    target_tz: TimeZone = "",
    unlocalized_tz: TimeZone = "",
) -> IntoFrameT:
    """Convert all datetime columns of a dataframe to target timezone.

    Ensures all datetime columns of a dataframe are timezone aware in the manner of datetime_localize:
    Columns without timezone information are first localized using the 'unlocalized_tz' parameter if it is provided, otherwise the localization will fall back to default.
    """
    df_localized = datetime_localize(df, target_tz=unlocalized_tz)
    nw_df = cast(nw.DataFrame, nw.from_native(df_localized))
    expression = _nw_expr_tz_convert(
        nw_df.schema,
        target_tz,
    )
    return nw_df.with_columns(expression).to_native()


def datelike_convert_timezone(
    df: IntoFrameT,
    target_tz: TimeZone = "",
    unlocalized_tz: TimeZone = "",
) -> IntoFrameT:
    """Convert all datelike columns of a dataframe to target timezone.

    Ensures all datetime columns of a dataframe are timezone aware in the manner of datetime_localize:
    Columns without timezone information are first localized using the 'unlocalized_tz' parameter if it is provided, otherwise the localization will fall back to default.
    """
    df_with_dt_cols = datelike_to_datetime(df)
    df_localized = datetime_localize(df_with_dt_cols, unlocalized_tz)
    return datetime_convert_timezone(df_localized, target_tz)


def datetime_to_utc(
    df: IntoFrameT,
    unlocalized_tz: TimeZone = "",
) -> IntoFrameT:
    """Convert datetime columns of a dataframe to UTC.

    If dates have no timezone information, the data is assumed to be in the default timezone (CET if not configured otherwise).
    """
    return datetime_convert_timezone(df, UTC, unlocalized_tz)


def datelike_to_utc(df: IntoFrameT, unlocalized_tz: TimeZone = "") -> IntoFrameT:
    """Convert all datelike columns of a dataframe to UTC."""
    df_localized = datelike_localize(df, target_tz=unlocalized_tz)
    return datetime_to_utc(df_localized)

    # chaining expresssions should have performance advantages
    # ... but does not work
    # all_expressions = [
    #     *_nw_expr_datelike_to_datetime(),
    #     *_nw_expr_tz_localize(nw_frame.schema, str(DEFAULT_TZ)),
    #     *_nw_expr_tz_convert(nw_frame.schema, 'UTC'),
    # ]
    # return nw_frame.with_columns(*all_expressions).to_native()


def validate_dates(
    df: IntoFrameT,
    date_columns: Iterable[str],
    throw_error: bool = False,
) -> bool:
    """Check that all expected date columns are defined, are time zone aware dates and in UTC."""
    nw_df = cast(nw.DataFrame, nw.from_native(df))
    if nw_df.is_empty():
        return True

    columns_exist = [d in nw_df.columns for d in date_columns]
    if all(columns_exist):
        ...
    elif throw_error:
        columns_not_found = set(date_columns) - set(nw_df.columns)
        raise ValueError(f"Expected date columns {columns_not_found} was not found.")
    else:
        return False

    dates_are_utc = [
        d in nw_df.select(ncs.datetime(time_zone="UTC")).columns for d in date_columns
    ]
    if all(dates_are_utc):
        ...
    elif throw_error:
        all_date_cols = nw_df.select(ncs.datetime()).schema
        utc_date_cols = nw_df.select(ncs.datetime(time_zone="UTC")).schema
        non_utc = set(all_date_cols.keys()) - set(utc_date_cols.keys())
        raise ValueError(
            f"Some provided date columns where not UTC: {non_utc}\n{all_date_cols}."
        )
    else:
        return False

    return all(columns_exist) and all(dates_are_utc)


def standardize_dates(
    df: nw.typing.IntoFrameT,
    time_unit: Literal["ns", "us", "ms"] = NW_DEFAULT_TIME_UNIT,
) -> nw.typing.IntoFrameT:
    """Ensure that all date columns conform to the same standards.

    * Same datatype --> nw.dt.Timestamp?
    * Time zone aware + UTC for storage.
    * Configurable max precision?

    Other questions/ideas include:
    * Pandas Period indexes are nice -> consider conversions?
    * Pendulum or other libraries?
    """
    as_utc = datelike_to_utc(df)
    return datetime_time_unit(as_utc, time_unit=time_unit)


def period_index(col: IntoSeriesT, freq: str) -> PeriodIndex:
    """Returns a period index for a date or datetime series."""
    dates = nw.from_native(col, series_only=True).to_pandas()
    return PeriodIndex(dates, freq=freq)
